import numpy as np
from torch.utils.data import Dataset

from ..transforms import functional as F
from .utils import sampling_rate_after_transform


__doctest_skip__ = ['*']


class WhiteNoise(Dataset):
    r"""White noise data set.

    The white noise is generated by numpy.random.standard_normal.

    * :attr:`duration` controls the duration of the noise signal
    * :attr:`sampling_rate` holds the sampling rate of the returned data
    * :attr:`mean` controls the mean of the underlying distribution
    * :attr:`stdev` controls the standard deviation of the underlying
      distribution
    * :attr:`transform` controls the input transform
    * :attr:`target_transform` controls the target transform

    As white noise has not really a sampling rate you can use the following
    attribute to change it instead of resampling:

    * :attr:`original_sampling_rate` controls the sampling rate of the data set

    Args:
        duration (float): duration of the noise signal in seconds
        sampling_rate (int, optional): sampling rate in Hz. Default: `44100`
        mean (float, optional): mean of underlying distribution. Default: `0`
        stdev (float, optional): standard deviation of underlying distribution.
            Default: `1`
        transform (callable, optional): function/transform applied on the
            signal. Default: `None`
        target_transform (callable, optional): function/transform applied on
            the target. Default: `None`

    Note:
        Even `WhiteNoise` has an infintely number of entries, its length is
        `1` as repeated calls of the same index return different signals.

    Example:
        >>> import sounddevice as sd
        >>> data = WhiteNoise(duration=1, sampling_rate=44100)
        >>> print(data)
        Dataset WhiteNoise
            Number of data points: Inf
            Signal length: 1s
            Sampling Rate: 44100Hz
            Label (str): noise type
        >>> signal, target = data[0]
        >>> target
        'white noise'
        >>> sd.play(signal.transpose(), data.sampling_rate)

    """

    def __init__(self, *, duration=1, sampling_rate=44100, mean=0, stdev=1,
                 transform=None, target_transform=None):
        super().__init__()
        self.duration = duration
        self.mean = mean
        self.stdev = stdev
        self.transform = transform
        self.target_transform = target_transform
        self.original_sampling_rate = sampling_rate

    @property
    def sampling_rate(self):
        return sampling_rate_after_transform(self)

    def __len__(self):
        # This has no meaningful __len__ as its actual length is `Inf`.
        # Return `1` to make it work with `random.choice` like operations.
        return 1

    def __getitem__(self, item):
        samples = int(np.ceil(self.duration * self.sampling_rate))
        signal = np.random.normal(loc=self.mean, scale=self.stdev,
                                  size=samples)
        signal = F.normalize(np.expand_dims(signal, axis=0))
        target = 'white noise'

        if self.transform is not None:
            signal = self.transform(signal)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return signal, target

    def __repr__(self):
        fmt_str = 'Dataset ' + self.__class__.__name__ + '\n'
        fmt_str += '    Number of data points: Inf\n'
        fmt_str += '    Signal length: {}s\n'.format(self.duration)
        if self.sampling_rate == self.original_sampling_rate:
            fmt_str += '    Sampling Rate: {}Hz\n'.format(self.sampling_rate)
        else:
            fmt_str += ('    Sampling Rate: {}Hz (original: {}Hz)\n'
                        .format(self.sampling_rate,
                                self.original_sampling_rate))
        fmt_str += '    Label (str): noise type\n'
        tmp1 = '    Transform: '
        tmp2 = self.transform.__repr__().replace('\n', '\n' + ' ' * len(tmp1))
        if self.transform:
            fmt_str += '{0}{1}\n'.format(tmp1, tmp2)
        tmp1 = '    Target Transform: '
        tmp2 = self.target_transform.__repr__().replace('\n',
                                                        '\n' + ' ' * len(tmp1))
        if self.target_transform:
            fmt_str += '{0}{1}'.format(tmp1, tmp2)
        return fmt_str



import os
import sys
from subprocess import check_output


# Import ------------------------------------------------------------------

# Relative source code path. Avoids `import audtorch`, which need package to be
# installed first.
sys.path.insert(0, os.path.abspath('..'))

# Ignore package dependencies during building the docs
autodoc_mock_imports = [
    'audiofile',
    'librosa',
    'numpy',
    'pandas',
    'resampy',
    'torch',
    'scipy',
    'tqdm',
    'tabulate',
]


# Project -----------------------------------------------------------------

project = 'audtorch'
copyright = '2019 audEERING GmbH'
author = ('Andreas Triantafyllopoulos, '
          'Stephan Huber, '
          'Johannes Wagner, '
          'Hagen Wierstorf')
# The x.y.z version read from tags
try:
    version = check_output(['git', 'describe', '--tags', '--always'])
    version = version.decode().strip()
except Exception:
    version = '<unknown>'
title = '{} Documentation'.format(project)


# General -----------------------------------------------------------------

master_doc = 'index'
extensions = []
source_suffix = '.rst'
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', '**.ipynb_checkpoints']
pygments_style = None
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',  # support for Google-style docstrings
    'sphinxcontrib.bibtex',
    'sphinxcontrib.katex',
]

napoleon_use_ivar = True  # List of class attributes
autodoc_inherit_docstrings = False  # disable docstring inheritance

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'numpy': ('https://docs.scipy.org/doc/numpy/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference/', None),
    'torch': ('https://pytorch.org/docs/stable/', None),
}

# HTML --------------------------------------------------------------------

html_theme = 'sphinx_rtd_theme'
html_theme_options = {
    'display_version': True,
    'logo_only': False,
}
html_title = title



import os
import json

import pandas as pd

from ..utils import flatten_list
from .base import AudioDataset
from .utils import safe_path


__doctest_skip__ = ['*']


class AudioSet(AudioDataset):
    r"""A large-scale dataset of manually annotated audio events.

    Open and publicly available data set of audio events from Google:
    https://research.google.com/audioset/

    License: CC BY 4.0

    The categories corresponding to an audio signal are returned as a list,
    starting with those included in the top hierarchy of the
    `AudioSet ontology`_, followed by those from the second hierarchy and then
    all other categories in a random order.

    The signals to be returned can be limited by excluding or including only
    certain categories. This is achieved by first including only the desired
    categories, estimating all its parent categories and then applying the
    exclusion.

    .. _AudioSet ontology: https://research.google.com/audioset/ontology/

    * :attr:`transform` controls the input transform
    * :attr:`target_transform` controls the target transform
    * :attr:`files` controls the audio files of the data set
    * :attr:`targets` controls the corresponding targets
    * :attr:`sampling_rate` holds the sampling rate of the returned data
    * :attr:`original_sampling_rate` holds the sampling rate of the audio files
      of the data set

    Args:
        root (str, optional): root directory of dataset.
            Default: `AudioSet.root`
        csv_file (str, optional): name of a CSV file located in `root`. Can be
            one of `balanced_train_segments.csv`,
            `unbalanced_train_segments.csv`, `eval_segments.csv`.
            Default: `balanced_train_segments.csv`
        include (list of str, optional): list of categories to include.
            If `None` all categories are included. Default: `None`
        exclude (list of str, optional): list of categories to exclude.
            If `None` no category is excluded. Default: `None`
        transform (callable, optional): function/transform applied on the
            signal. Default: `None`
        target_transform (callable, optional): function/transform applied on
            the target. Default: `None`

    `AudioSet ontology`_ categories of the two top hierarchies:

    .. code-block:: none

        Human sounds            Animal                   Music
        |-Human voice           |-Domestic animals, pets |-Musical instrument
        |-Whistling             |-Livestock, farm        |-Music genre
        |-Respiratory sounds    | animals, working       |-Musical concepts
        |-Human locomotion      | animals                |-Music role
        |-Digestive             \-Wild animals           \-Music mood
        |-Hands
        |-Heart sounds,         Sounds of things         Natural sounds
        | heartbeat             |-Vehicle                |-Wind
        |-Otoacoustic emission  |-Engine                 |-Thunderstorm
        \-Human group actions   |-Domestic sounds,       |-Water
                                | home sounds            \-Fire
        Source-ambiguous sounds |-Bell
        |-Generic impact sounds |-Alarm                  Channel, environment
        |-Surface contact       |-Mechanisms             and background
        |-Deformable shell      |-Tools                  |-Acoustic environment
        |-Onomatopoeia          |-Explosion              |-Noise
        |-Silence               |-Wood                   \-Sound reproduction
        \-Other sourceless      |-Glass
                                |-Liquid
                                |-Miscellaneous sources
                                \-Specific impact sounds

    Warning:
        Some of the recordings in `AudioSet` were captured with `mono` and
        others with `stereo` input. The user must be careful to handle this,
        e.g. using a transform to adjust number of channels.

    Example:
        >>> import sounddevice as sd
        >>> data = AudioSet(root='/data/AudioSet', include=['Thunderstorm'])
        >>> print(data)
        Dataset AudioSet
            Number of data points: 73
            Root Location: /data/AudioSet
            Sampling Rate: 16000Hz
            CSV file: balanced_train_segments.csv
            Included categories: ['Thunderstorm']
        >>> signal, target = data[4]
        >>> target
        ['Natural sounds', 'Thunderstorm', 'Water', 'Rain', 'Thunder']
        >>> sd.play(signal.transpose(), data.sampling_rate)

    """

    # Categories of the two top hieararchies of AudioSet
    # https://research.google.com/audioset/ontology/
    categories = {
        'Human sounds': [
            'Human voice', 'Whistling', 'Respiratory sounds',
            'Human locomotion', 'Digestive', 'Hands',
            'Heart sounds, heartbeat', 'Otoacoustic emission',
            'Human group actions'],
        'Source-ambiguous sounds': [
            'Generic impact sounds', 'Surface contact', 'Deformable shell',
            'Onomatopoeia', 'Silence', 'Other sourceless'],
        'Animal': [
            'Domestic animals, pets',
            'Livestock, farm animals, working animals', 'Wild animals'],
        'Sounds of things': [
            'Vehicle', 'Engine', 'Domestic sounds, home sounds', 'Bell',
            'Alarm', 'Mechanisms', 'Tools', 'Explosion', 'Wood', 'Glass',
            'Liquid', 'Miscellaneous sources', 'Specific impact sounds'],
        'Music': [
            'Musical instrument', 'Music genre', 'Musical concepts',
            'Music role', 'Music mood'],
        'Natural sounds': [
            'Wind', 'Thunderstorm', 'Water', 'Fire'],
        'Channel, environment and background': [
            'Acoustic environment', 'Noise', 'Sound reproduction']
    }

    def __init__(self, root, csv_file='balanced_train_segments.csv',
                 sampling_rate=16000, include=None, exclude=None,
                 transform=None, target_transform=None):
        root = safe_path(root)
        # Allow only official CSV files as no audio paths are defined otherwise
        assert csv_file in ['eval_segments.csv', 'balanced_train_segments.csv',
                            'unbalanced_train_segments.csv']
        csv_file = os.path.join(root, csv_file)

        # Load complete ontology
        with open(os.path.join(root, 'ontology.json')) as fp:
            self.ontology = json.load(fp)

        # Get the desired filenames and categories
        df = pd.read_csv(csv_file, skiprows=2, sep=', ', engine='python')
        df = self._filename_and_ids(df)
        if include is not None:
            df = self._filter_by_categories(df, include)
        df['ids'] = df['ids'].map(self._add_parent_ids)
        if exclude is not None:
            df = self._filter_by_categories(df, exclude, exclude_mode=True)
        categories = df['ids'].map(self._convert_ids_to_categories)

        audio_folder = os.path.splitext(os.path.basename(csv_file))[0]
        file_root = os.path.join(root, audio_folder)
        files = [os.path.join(file_root, f) for f in df['filename']]

        super().__init__(root, files, targets=categories, sampling_rate=16000,
                         transform=transform,
                         target_transform=target_transform)
        self.csv_file = csv_file
        self.include = include
        self.exclude = exclude

    def _filename_and_ids(self, df):
        r"""Return data frame with filenames and IDs.

        Args:
            df (pandas.DataFrame): data frame as read in from the CSV file

        Results:
            pandas.DataFrame: data frame with columns `filename` and `ids`

        """
        df.rename(columns={'positive_labels': 'ids'}, inplace=True)
        # Translate labels from "label1,label2" to [label1, label2]
        df['ids'] = [label.strip('\"').split(',') for label in df['ids']]
        # Insert filename
        df['filename'] = (df['# YTID']
                          + '_'
                          + ['{:.3f}'.format(x) for x in df['start_seconds']]
                          + '.wav')
        return df[['filename', 'ids']]

    def _add_parent_ids(self, child_ids):
        r"""Add all parent IDs to the list of given child IDs.

        Args:
            child_ids (list of str): child IDs

        Return:
            list of str: list of child and parent IDs

        """
        ids = child_ids
        for id in child_ids:
            ids += [x['id'] for x in self.ontology if id in x['child_ids']]
        # Remove duplicates
        return list(set(ids))

    def _convert_ids_to_categories(self, ids):
        r"""Convert list of ids to sorted list of categories.

        Args:
            ids (list of str): list of IDs

        Returns:
            list of str: list of sorted categories

        """
        # Convert IDs to categories
        categories = []
        for id in ids:
            categories += [x['name'] for x in self.ontology
                           if x['id'] == id]
        # Order categories after the first two top ontologies
        order = []
        first_hierarchy = self.categories.keys()
        second_hierarchy = flatten_list(list(self.categories.values()))
        for cat in categories:
            if cat in first_hierarchy:
                order += [0]
            elif cat in second_hierarchy:
                order += [1]
            else:
                order += [2]
        # Sort list `categories` by the list `order`
        categories = [cat for _, cat in sorted(zip(order, categories))]
        return categories

    def _filter_by_categories(self, df, categories, exclude_mode=False):
        r"""Return data frame containing only specified categories.

        Args:
            df (pandas.DataFrame): data frame containing the columns `ids`
            categories (list of str): list of categories to include or exclude
            exclude_mode (bool, optional): if `False` the specified categories
                should be included in the data frame, otherwise excluded.
                Default: `False`

        Returns:
            pandas.DataFrame: data frame containing only the desired categories

        """
        ids = self._ids_for_categories(categories)
        if exclude_mode:
            # Remove rows that have an intersection of actual and desired IDs
            df = df[[False if set(row['ids']) & set(ids) else True
                     for _, row in df.iterrows()]]
        else:
            # Include rows that have an intersection of actual and desired IDs
            df = df[[True if set(row['ids']) & set(ids) else False
                     for _, row in df.iterrows()]]
        df = df.reset_index(drop=True)
        return df

    def _ids_for_categories(self, categories):
        r"""All IDs and child IDs for a given set of categories.

        Args:
            categories (list of str): list of categories

        Returns:
            list: list of IDs

        """
        ids = []
        category_ids = \
            [x['id'] for x in self.ontology if x['name'] in categories]
        for category_id in category_ids:
            ids += self._subcategory_ids(category_id)
        # Remove duplicates
        return list(set(ids))

    def _subcategory_ids(self, parent_id):
        r"""Recursively identify all IDs of a given category.

        Args:
            parent_id (unicode str): ID of parent category

        Returns:
            list: list of all children IDs and the parent ID

        """
        id_list = [parent_id]
        child_ids = \
            [x['child_ids'] for x in self.ontology if x['id'] == parent_id]
        child_ids = flatten_list(child_ids)
        # Add all subcategories
        for child_id in child_ids:
            id_list += self._subcategory_ids(child_id)
        return id_list

    def extra_repr(self):
        fmt_str = '    CSV file: {}\n'.format(os.path.basename(self.csv_file))
        if self.include:
            fmt_str += '    Included categories: {}\n'.format(self.include)
        if self.exclude:
            fmt_str += '    Excluded categories: {}\n'.format(self.exclude)
        return fmt_str



import torch
from torch.nn.utils.rnn import pad_sequence


class Collation(object):
    r"""Abstract interface for collation classes.

    All other collation classes should subclass it. All subclasses should
    override ``__call__``, that executes the actual collate function.

    """

    def __init__(self):
        super().__init__()

    def __call__(self, batch):
        raise NotImplementedError("This is an abstract interface for "
                                  "modularizing collate functions."
                                  "Please use one of its subclasses.")


class Seq2Seq(Collation):
    r"""Pads mini-batches to longest contained sequence for seq2seq-purposes.

    This class pads features and targets to the largest sequence in the batch.
    Before padding, length information are extracted from them.

    Note:
        The tensors can be sorted in descending order of features' lengths
        by enabling :attr:`sort_sequences`. Thereby the requirements of
        :py:func:`torch.nn.utils.rnn.pack_padded_sequence`
        are anticipated, which is used by recurrent layers.

    * :attr:`sequence_dimensions` holds dimension of sequence in features
      and targets
    * :attr:`batch_first` controls output shape of features and targets
    * :attr:`pad_values` controls values to pad features (targets) with
    * :attr:`sort_sequences` controls if sequences are sorted in
      descending order of `features`' lengths

    Args:
        sequence_dimensions (list of ints): indices representing dimension of
            sequence in feature and target tensors.
            Position `0` represents sequence dimension of `features`,
            position `1` represents sequence dimension of `targets`.
            Negative indexing is permitted
        batch_first (bool or None, optional): determines output shape of
            collate function. If `None`, original shape of
            `features` and `targets` is kept with dimension of `batch size`
            prepended. See Shape for more information.
            Default: `None`
        pad_values (list, optional): values to pad shorter sequences with.
            Position `0` represents value of `features`,
            position `1` represents value of `targets`. Default: `[0, 0]`
        sort_sequences (bool, optional): option whether to sort sequences
            in descending order of `features`' lengths. Default: `True`

    Shape:
        - Input: :math:`(*, S, *)`, where :math:`*` can be any number
          of further dimensions except :math:`N` which is the batch size,
          and where :math:`S` is the sequence dimension.
        - Output:

          - `features`:

            - :math:`(N, *, S, *)` if :attr:`batch_first` is `None`,
              i.e. the original input shape with :math:`N` prepended
              which is the batch size
            - :math:`(N, S, *, *)` if :attr:`batch_first` is `True`
            - :math:`(S, N, *, *)` if :attr:`batch_first` is `False`

          - `feats_lengths`: :math:`(N,)`

          - `targets`: analogous to `features`

          - `tgt_lengths`: analogous to `feats_lengths`

    Example:
        >>> # data format: FS = (feature dimension, sequence dimension)
        >>> batch = [[torch.zeros(161, 108), torch.zeros(10)],
        ...          [torch.zeros(161, 223), torch.zeros(12)]]
        >>> collate_fn = Seq2Seq([-1, -1], batch_first=None)
        >>> features = collate_fn(batch)[0]
        >>> list(features.shape)
        [2, 161, 223]

    """

    def __init__(self, sequence_dimensions, batch_first=None,
                 pad_values=[0, 0], sort_sequences=True):

        self.sequence_dimensions = sequence_dimensions
        self.batch_first = batch_first
        self.pad_values = pad_values
        self.sort_sequences = sort_sequences

    def __call__(self, batch):
        r"""Collate and pad sequences of mini-batch.

        The output tensor is augmented by the dimension of `batch_size`.

        Args:
            batch (list of tuples): contains all samples of a batch.
                Each sample is represented by a tuple (`features`, `targets`)
                which is returned by data set's __getitem__ method

        Returns:
            torch.tensors: `features`, `feature lengths`, `targets`
                and `target lengths` in data format according to
                :attr:`batch_first`.

        """
        features = [torch.as_tensor(sample[0]) for sample in batch]
        features, feats_lengths, sorted_indices = _collate_sequences(
            features, self.sequence_dimensions[0], self.pad_values[0],
            self.batch_first, self.sort_sequences, [])

        targets = [torch.as_tensor(sample[1]) for sample in batch]
        targets, tgt_lengths, _ = _collate_sequences(
            targets, self.sequence_dimensions[1], self.pad_values[1],
            self.batch_first, self.sort_sequences, sorted_indices)

        return features, feats_lengths, targets, tgt_lengths


def _collate_sequences(sequences, sequence_dimension, pad_value,
                       batch_first, sort_sequences=True, sorted_indices=[]):
    r"""Collate and pad sequences.

    Args:
        sequences (list of torch.tensors): contains all samples of a batch
        sequence_dimension (int): index representing dimension of sequence
            in tensors
        batch_first (bool or None): determines output shape of tensors
        pad_value (float): value to pad shorter sequences with.
        sort_sequences (bool, optional): option whether to sort `sequences`.
            Default: `True`
        sorted_indices (list of ints, optional): indices to sort sequences
            and their lengths in descending order with. Default: `[]`

    Returns:
        tuple:

        * torch.Tensor: data of sequences in format
          according to :attr:`batch_first`, list of sorted indices
        * torch.IntTensor: lengths of sequences
        * list of int: indices of sequences sorted in descending order
          of their lengths

    """
    # handle negative indexing
    sequence_dimension = len(sequences[0].shape) + sequence_dimension \
        if sequence_dimension < 0 else sequence_dimension

    # input to `pad_sequence` requires shape `(S, *)`
    # swap sequence-dimension to the front
    sequences = [t.transpose(0, sequence_dimension) for t in sequences]

    # extract lengths (first dimension of permuted tensor)
    lengths = [t.shape[0] for t in sequences]

    if sort_sequences:
        # sort sequences and lengths in descending order of `lengths`
        if not sorted_indices:
            sorted_indices = sorted(
                range(len(lengths)), key=lambda i: lengths[i], reverse=True)

        sequences = [sequences[idx] for idx in sorted_indices]
        lengths = [lengths[idx] for idx in sorted_indices]

    # pad sequences
    sequences = pad_sequence(
        sequences=sequences,
        batch_first=batch_first if batch_first is not None else True,
        padding_value=pad_value)

    if batch_first is None:  # recover input data format
        sequences = sequences.transpose(1, sequence_dimension + 1)

    else:  # recover order of "*"-dimensions
        permuted = list(range(len(sequences.shape)))
        if sequence_dimension >= 2:
            permuted.insert(
                2, permuted.pop(sequence_dimension + 1))
        sequences = sequences.permute(permuted)

    lengths = torch.IntTensor(lengths)

    return sequences, lengths, sorted_indices



import torch.nn as nn
import torch.nn.functional as F


class EnergyConservingLoss(nn.L1Loss):
    r"""Energy conserving loss.

    A two term loss that enforces energy conservation after
    :cite:`Rethage2018`.

    The loss can be described as:

    .. math::
        \ell(x, y, m) = L = \{l_1,\dots,l_N\}^\top, \quad
        l_n = |x_n - y_n| + |b_n - \hat{b_n}|,

    where :math:`N` is the batch size. If reduction is not ``'none'``, then:

    .. math::
        \ell(x, y, m) =
        \begin{cases}
            \operatorname{mean}(L), & \text{if reduction} = \text{'mean';}\\
            \operatorname{sum}(L),  & \text{if reduction} = \text{'sum'.}
        \end{cases}

    :math:`x` is the input signal (estimated target), :math:`y` the target
    signal, :math:`m` the mixture signal, :math:`b` the background signal given
    by :math:`b = m - y`, and :math:`\hat{b}` the estimated background signal
    given by :math:`\hat{b} = m - x`.

    Args:
        reduction (string, optional): specifies the reduction to apply to the
            output: 'none' | 'mean' | 'sum'.
            'none': no reduction will be applied, 'mean': the sum of the output
            will be divided by the number of elements in the output, 'sum': the
            output will be summed.

    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional
          dimensions
        - Target: :math:`(N, *)`, same shape as the input
        - Mixture: :math:`(N, *)`, same shape as the input
        - Output: scalar. If reduction is ``'none'``, then :math:`(N, *)`, same
          shape as the input

    Examples:
        >>> import torch
        >>> _ = torch.manual_seed(0)
        >>> loss = EnergyConservingLoss()
        >>> input = torch.randn(3, 5, requires_grad=True)
        >>> target = torch.randn(3, 5)
        >>> mixture = torch.randn(3, 5)
        >>> loss(input, target, mixture)
        tensor(2.1352, grad_fn=<AddBackward0>)

    """
    def __init__(self, reduction='mean'):
        super().__init__(None, None, reduction)

    def forward(self, y_predicted, y, x):
        noise = x - y
        noise_predicted = x - y_predicted
        return (F.l1_loss(y_predicted, y, reduction=self.reduction)
                + F.l1_loss(noise_predicted, noise, reduction=self.reduction))



import os
import glob
import shutil

import pandas as pd

from .utils import (download_url_list, extract_archive, safe_path)
from ..utils import run_worker_threads
from .base import PandasDataset


__doctest_skip__ = ['*']


class LibriSpeech(PandasDataset):
    r"""`LibriSpeech` speech data set.

    Open and publicly available data set of voices from OpenSLR:
    http://www.openslr.org/12/

    License: CC BY 4.0.

    `LibriSpeech` contains several hundred hours of English speech
    with corresponding transcriptions in capital letters without punctuation.

    It is split into different subsets according to WER-level achieved when
    performing speech recognition on the speakers. The subsets are:
    `train-clean-100`, `train-clean-360`, `train-other-500` `dev-clean`,
    `dev-other`, `test-clean`, `test-other`

    * :attr:`root` holds the data set's location
    * :attr:`transform` controls the input transform
    * :attr:`target_transform` controls the target transform
    * :attr:`files` controls the audio files of the data set
    * :attr:`labels` controls the corresponding labels
    * :attr:`sampling_rate` holds the sampling rate of data set

    In addition, the following class attributes are available

    * :attr:`all_sets` holds the names of the different pre-defined sets
    * :attr:`urls` holds the download links of the different sets

    Args:
        root (str): root directory of data set
        sets (str or list, optional): desired sets of `LibriSpeech`.
            Mutually exclusive with :attr:`dataframe`.
            Default: `None`
        dataframe (pandas.DataFrame, optional): pandas data frame containing
            columns `audio_path` (relative to root) and `transcription`.
            It can be used to pre-select files based on meta information,
            e.g. sequence length. Mutually exclusive with :attr:`sets`.
            Default: `None`
        transform (callable, optional): function/transform applied on
            the signal. Default: `None`
        target_transform (callable, optional): function/transform applied on
            the target. Default: `None`
        download (bool, optional): download data set to root directory
            if not present. Default: `False`

    Example:
        >>> import sounddevice as sd
        >>> data = LibriSpeech(root='/data/LibriSpeech', sets='dev-clean')
        >>> print(data)
        Dataset LibriSpeech
            Number of data points: 2703
            Root Location: /data/LibriSpeech
            Sampling Rate: 16000Hz
            Sets: dev-clean
        >>> signal, label = data[8]
        >>> label
        AS FOR ETCHINGS THEY ARE OF TWO KINDS BRITISH AND FOREIGN
        >>> sd.play(signal.transpose(), data.sampling_rate)

    """

    all_sets = ['train-clean-100', 'train-clean-360', 'train-other-500',
                'dev-clean', 'dev-other', 'test-clean', 'test-other']
    urls = {
        "train-clean-100":
            'https://openslr.org/resources/12/train-clean-100.tar.gz',
        "train-clean-360":
            'https://openslr.org/resources/12/train-clean-360.tar.gz',
        "train-other-500":
            'https://openslr.org/resources/12/train-other-500.tar.gz',
        "dev-clean":
            'https://openslr.org/resources/12/dev-clean.tar.gz',
        "dev-other":
            'https://openslr.org/resources/12/dev-other.tar.gz',
        "test-clean":
            'https://openslr.org/resources/12/test-clean.tar.gz',
        "test-other":
            'https://openslr.org/resources/12/test-other.tar.gz'}
    _transcription = 'transcription'
    _audio_path = 'audio_path'

    def __init__(self, root, *, sets=None, dataframe=None,
                 transform=None, target_transform=None, download=False):

        self.root = safe_path(root)

        if isinstance(sets, str):
            sets = [sets]
        if dataframe is None and sets is None:
            self.sets = self.all_sets
        elif dataframe is None:
            assert set(sets) <= set(self.all_sets)
            self.sets = sets
        elif dataframe is not None:
            self.sets = None
        else:
            raise ValueError('Either `sets` or `dataframe` can be specified.')

        if download:  # data not available
            self._download()

        if not self._check_exists():
            raise RuntimeError('Requested sets of data set not found.')

        if dataframe is None:
            files = self._get_files()
            dataframe = self._create_dataframe(files)

        super().__init__(
            root=self.root,
            sampling_rate=16000,
            df=dataframe,
            column_filename=self._audio_path,
            column_labels=self._transcription,
            transform=transform,
            target_transform=target_transform)

    def _check_exists(self):
        return all([os.path.exists(os.path.join(self.root, s))
                    for s in self.sets])

    def _download(self):
        absent_sets = [s for s in self.sets
                       if not os.path.exists(os.path.join(self.root, s))]
        if not absent_sets:
            return

        out_path = os.path.join(self.root, "tmp")
        if not os.path.exists(out_path):
            os.makedirs(out_path)

        urls = [self.urls[s] for s in absent_sets]
        filenames = download_url_list(urls, out_path, num_workers=0)
        for filename in filenames:
            extract_archive(os.path.join(out_path, filename),
                            out_path=out_path,
                            remove_finished=True)
        contents = glob.glob(os.path.join(out_path, 'LibriSpeech/*'))
        for f in contents:
            shutil.move(f, self.root)
        os.rmdir(os.path.join(out_path, "LibriSpeech"))
        os.rmdir(out_path)

    def _get_files(self):
        files = []
        for set in self.sets:
            path_to_files = os.path.join(self.root, set, '**/**/*.flac')
            files += glob.glob(path_to_files)
        return files

    @classmethod
    def _create_dataframe(cls, files):

        def _create_df_per_txt(txt_file):
            _set = txt_file.rsplit('/', 4)[-4]
            df_per_txt = pd.read_csv(txt_file, names=[cls._audio_path])

            # split content once with delimiter
            df_per_txt[[cls._audio_path, cls._transcription]] = \
                df_per_txt[cls._audio_path].str.split(" ", 1, expand=True)

            # compose audio paths relative to root
            df_per_txt[cls._audio_path] = df_per_txt[cls._audio_path].apply(
                _compose_relative_audio_paths(_set))
            return df_per_txt

        def _compose_relative_audio_paths(_set):
            return lambda row: os.path.join(
                _set, *row.split('-')[:-1], row + '.flac')

        # get absolute paths to txt files from audio files
        txt_files = sorted(list(set(
            [f.rsplit('-', 1)[0] + '.trans.txt' for f in files])))
        args = [(f, ) for f in txt_files]
        dataframes = run_worker_threads(12, _create_df_per_txt, args)
        return pd.concat(dataframes, ignore_index=True)

    def extra_repr(self):
        if self.sets is not None:
            fmt_str = '    Sets: {}\n'.format(", ".join(self.sets))
        return fmt_str



from . import functional as F


class PearsonR(object):
    r"""Computes Pearson Correlation Coefficient.

    Pearson Correlation Coefficient (also known as Linear Correlation
    Coefficient or Pearson's :math:`\rho`) is computed as:

    .. math::

        \rho = \frac {E[(X-\mu_X)(Y-\mu_Y)]} {\sigma_X\sigma_Y}

    If inputs are vectors, computes Pearson's :math:`\rho` between the two
    of them. If inputs are multi-dimensional arrays, computes Pearson's
    :math:`\rho` along the first or last input dimension according to the
    `batch_first` argument, returns a **torch.Tensor** as output,
    and optionally reduces it according to the `reduction` argument.

    Args:
        reduction (string, optional): specifies the reduction to apply to the
            output: 'none' | 'mean' | 'sum'.
            'none': no reduction will be applied, 'mean': the sum of the output
            will be divided by the number of elements in the output, 'sum': the
            output will be summed. Default: 'mean'
        batch_first (bool, optional): controls if batch dimension is first.
            Default: `True`

    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional
          dimensions
        - Target: :math:`(N, *)`, same shape as the input
        - Output: scalar. If reduction is ``'none'``, then :math:`(N, 1)`

    Example:
        >>> import torch
        >>> _ = torch.manual_seed(0)
        >>> metric = PearsonR()
        >>> input = torch.rand(3, 5)
        >>> target = torch.rand(3, 5)
        >>> metric(input, target)
        tensor(0.1220)

    """
    def __init__(self, reduction='mean', batch_first=True):
        self.reduction = reduction
        self.batch_first = batch_first

    def __call__(self, x, y):
        r = F.pearsonr(x, y, self.batch_first)
        if self.reduction == 'mean':
            r = r.mean()
        elif self.reduction == 'sum':
            r = r.sum()
        return r


class ConcordanceCC(object):
    r"""Computes Concordance Correlation Coefficient (CCC).

    CCC is computed as:

    .. math::

        \rho_c = \frac {2\rho\sigma_X\sigma_Y} {\sigma_X\sigma_X +
        \sigma_Y\sigma_Y + (\mu_X - \mu_Y)^2}

    where :math:`\rho` is Pearson Correlation Coefficient, :math:`\sigma_X`,
    :math:`\sigma_Y` are the standard deviation, and :math:`\mu_X`,
    :math:`\mu_Y` the mean values of :math:`X` and :math:`Y` accordingly.

    If inputs are vectors, computes CCC between the two of them. If inputs
    are multi-dimensional arrays, computes CCC along the first or last input
    dimension according to the `batch_first` argument, returns a
    **torch.Tensor** as output, and optionally reduces it according to the
    `reduction` argument.

    Args:
        reduction (string, optional): specifies the reduction to apply to the
            output: 'none' | 'mean' | 'sum'.
            'none': no reduction will be applied, 'mean': the sum of the output
            will be divided by the number of elements in the output, 'sum': the
            output will be summed. Default: 'mean'
        batch_first (bool, optional): controls if batch dimension is first.
            Default: `True`

    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional
          dimensions
        - Target: :math:`(N, *)`, same shape as the input
        - Output: scalar. If reduction is ``'none'``, then :math:`(N, 1)`

    Example:
        >>> import torch
        >>> _ = torch.manual_seed(0)
        >>> metric = ConcordanceCC()
        >>> input = torch.rand(3, 5)
        >>> target = torch.rand(3, 5)
        >>> metric(input, target)
        tensor(0.0014)

    """
    def __init__(self, reduction='mean', batch_first=True):
        self.reduction = reduction
        self.batch_first = batch_first

    def __call__(self, x, y):
        r = F.concordance_cc(x, y, self.batch_first)
        if self.reduction == 'mean':
            r = r.mean()
        elif self.reduction == 'sum':
            r = r.sum()
        return r



import pytest
import numpy as np

import audtorch as at


xfail = pytest.mark.xfail


@pytest.mark.parametrize('nested_list,expected_list', [
    ([1, 2, 3, [4], [], [[[[[[[[[5]]]]]]]]]], [1, 2, 3, 4, 5]),
    ([[1, 2], 3], [1, 2, 3]),
    ([1, 2, 3], [1, 2, 3]),
])
def test_flatten_list(nested_list, expected_list):
    flattened_list = at.utils.flatten_list(nested_list)
    assert flattened_list == expected_list


@pytest.mark.parametrize('input,tuple_len,expected_output', [
    ('aa', 2, ('a', 'a')),
    (2, 1, (2,)),
    (1, 3, (1, 1, 1)),
    ((1, (1, 2)), 2, (1, (1, 2))),
    ([1, 2], 2, (1, 2)),
    pytest.param([1], 2, [], marks=xfail(raises=ValueError)),
    pytest.param([], 2, [], marks=xfail(raises=ValueError)),
])
def test_to_tuple(input, tuple_len, expected_output):
    output = at.utils.to_tuple(input, tuple_len=tuple_len)
    assert output == expected_output


@pytest.mark.parametrize('input,expected_output', [
    (np.array([[2, 2]]), 8),
])
def test_energy(input, expected_output):
    output = at.utils.energy(input)
    assert output == expected_output


@pytest.mark.parametrize('input,expected_output', [
    (np.array([[2, 2]]), 4),
])
def test_power(input, expected_output):
    output = at.utils.power(input)
    assert output == expected_output


@pytest.mark.parametrize('n_workers,task_fun,params', [
    (3, lambda x, n: x ** n, [(2, n) for n in range(10)]),
])
def test_run_worker_threads(n_workers, task_fun, params):
    list1 = at.utils.run_worker_threads(n_workers, task_fun, params)
    list2 = [task_fun(*p) for p in params]
    assert len(list1) == len(list2) and list1 == list2



import random

import numpy as np
from torch.utils.data import Dataset

from .base import _include_repr
from .utils import ensure_same_sampling_rate


__doctest_skip__ = ['*']


class SpeechNoiseMix(Dataset):
    r"""Mix speech and noise with speech as target.

    Add noise to each speech sample from the provided data by a
    mix transform. Return the mix as input and the speech signal as
    corresponding target. In addition, allow to replace randomly some of the
    mixes by noise as input and silence as output. This helps to train a speech
    enhancement algorithm to deal with background noise only as input signal
    :cite:`Rethage2018`.

    * :attr:`speech_dataset` controls the speech data set
    * :attr:`mix_transform` controls the transform that adds noise
    * :attr:`transform` controls the transform applied on the mix
    * :attr:`target_transform` controls the transform applied on the target
      clean speech
    * :attr:`joint_transform` controls the transform applied jointly on the
      mixture an the target clean speech
    * :attr:`percentage_silence` controls the amount of noise-silent data
      augmentation

    Args:
        speech_dataset (Dataset): speech data set
        mix_transform (callable): function/transform that can augment a signal
            with noise
        transform (callable, optional): function/transform applied on the
            speech-noise-mixture (input) only. Default; `None`
        target_transform (callable, optional): function/transform applied
            on the speech (target) only. Default: `None`
        joint_transform (callable, optional): function/transform applied
            on the mixtue (input) and speech (target) simultaneously. If the
            transform includes randomization it is applied with the same random
            parameter during both calls
        percentage_silence (float, optional): value between `0` and `1`, which
            controls the percentage of randomly inserted noise input, silent
            target pairs. Default: `0`

    Examples:
        >>> import sounddevice as sd
        >>> from audtorch import datasets, transforms
        >>> noise = datasets.WhiteNoise(duration=10, sampling_rate=48000)
        >>> mix = transforms.RandomAdditiveMix(noise)
        >>> normalize = transforms.Normalize()
        >>> speech = datasets.MozillaCommonVoice(root='/data/MozillaCommonVoice/cv_corpus_v1')
        >>> data = SpeechNoiseMix(speech, mix, transform=normalize)
        >>> noisy, clean = data[0]
        >>> sd.play(noisy.transpose(), data.sampling_rate)

    """  # noqa: E501
    def __init__(self, speech_dataset, mix_transform, *, transform=None,
                 target_transform=None, joint_transform=None,
                 percentage_silence=0):
        super().__init__()
        self.speech_dataset = speech_dataset
        self.mix_transform = mix_transform
        self.transform = transform
        self.target_transform = target_transform
        self.joint_transform = joint_transform
        self.percentage_silence = percentage_silence

        if not (0 <= self.percentage_silence <= 1):
            raise ValueError('`percentage_silence` needs to be in [0, 1]`')

        if hasattr(mix_transform, 'dataset'):
            ensure_same_sampling_rate([speech_dataset, mix_transform.dataset])

    def __len__(self):
        return len(self.speech_dataset)

    def __getitem__(self, item):
        # [0] ensures that we get only data, no targets
        speech = self.speech_dataset[item][0]
        # Randomly add (noise, silence) as (input, target)
        if random.random() < self.percentage_silence:
            speech = np.zeros(speech.shape)
        mixture = self.mix_transform(speech)

        if self.joint_transform is not None:
            randomness = getattr(self.joint_transform, 'fix_randomization',
                                 None)
            mixture = self.joint_transform(mixture)
            if randomness is not None:
                self.joint_transform.fix_randomization = True
            speech = self.joint_transform(speech)
            if randomness is not None:
                self.joint_transform.fix_randomization = randomness

        if self.transform is not None:
            mixture = self.transform(mixture)

        if self.target_transform is not None:
            speech = self.target_transform(speech)

        # input, target
        return mixture, speech

    @property
    def sampling_rate(self):
        return self.speech_dataset.sampling_rate

    def __repr__(self):
        speech_dataset_name = self.speech_dataset.__class__.__name__
        fmt_str = 'Dataset {}\n'.format(self.__class__.__name__)
        fmt_str += '    Number of data points: {}\n'.format(self.__len__())
        fmt_str += '    Speech dataset: {}\n'.format(speech_dataset_name)
        fmt_str += '    Sampling rate: {}Hz\n'.format(self.sampling_rate)
        if self.percentage_silence > 0:
            fmt_str += ('    Silence augmentation: {:.0f}%\n'
                        .format(100 * self.percentage_silence))
        fmt_str += '    Labels: speech signal\n'
        fmt_str += _include_repr('Mixing Transform', self.mix_transform)
        if self.transform:
            fmt_str += _include_repr('Transform', self.transform)
        if self.target_transform:
            fmt_str += _include_repr('Target Transform', self.target_transform)
        if self.joint_transform:
            fmt_str += _include_repr('Joint Transform', self.joint_transform)
        return fmt_str



from .audio_set import *
from .base import *
from .libri_speech import *
from .mixture import *
from .mozilla_common_voice import *
from .voxceleb1 import *
from .white_noise import *
from .utils import *



import numpy as np
from torch.utils.data import Sampler


class BucketSampler(Sampler):
    r"""Creates batches from ordered data sets.

    This sampler iterates over the data sets of `concat_dataset`
    and samples sequentially from them.
    Samples of each batch deliberately originate solely from the same data set.
    Only when the current data set is exhausted, the next data set
    is sampled from. In other words,
    samples from different buckets are never mixed.

    In each epoch `num_batches` batches of size `batch_sizes`
    are extracted from each data set.
    If the requested number of batches cannot be extracted from a data set,
    only its available batches are queued.
    By default, the data sets (and thus their batches) are iterated over
    in increasing order of their data set id.

    Note:
        The information in :attr:`batch_sizes` and :attr:`num_batches`
        refer to :attr:`datasets` at the same index
        independently of :attr:`permuted_order`.

    Simple Use Case: "Train on data with increasing sequence length"

    ======================= ==================================
    bucket_id:              [0,     1,      2,     ...  end  ]
    batch_sizes:            [32,    16,     8,     ...  2    ]
    num_batches:            [None,  None,   None,  ...  None ]
    ======================= ==================================

    Result:
    "Extract all batches (`None`) from all data sets,
    all of different batch size, and queue them
    in increasing order of their data set id"

    * :attr:`batch_sizes` controls batch size for each data set
    * :attr:`num_batches` controls number of batches to extract
      from each data set
    * :attr:`permuted_order` controls if order in which data sets are iterated
      over is permuted or in which specific order iteration is permuted
    * :attr:`shuffle_each_bucket` controls if each data set is shuffled
    * :attr:`drop_last` controls whether to drop last samples of a bucket
      which cannot form a mini-batch

    Args:
        concat_dataset (torch.utils.data.ConcatDataset): ordered concatenated
            data set
        batch_sizes (list): batch sizes per data set. Permissible values are
            unsigned integers
        num_batches (list or None, optional): number of batches per data set.
            Permissible values are non-negative integers and None.
            If None, then as many batches are extracted as data set provides.
            Default: `None`
        permuted_order (bool or list, optional): option whether to permute the
            order of data set ids in which the respective data set's batches
            are queued. If True (False), data set ids are (not) shuffled.
            Besides, a customized list of permuted data set ids can be
            specified. Default: `False`
        shuffle_each_bucket (bool, optional): option whether to shuffle samples
            in each data set. Recommended to set to True. Default: `True`
        drop_last (bool, optional): controls whether the last samples of a
            bucket which cannot form a mini-batch should be dropped.
            Default: `False`

    Example:
        >>> import torch
        >>> from torch.utils.data import (TensorDataset, ConcatDataset)
        >>> from audtorch.datasets.utils import defined_split
        >>> data = TensorDataset(torch.randn(100))
        >>> lengths = np.random.randint(0, 890, (100,))
        >>> split_func = buckets_of_even_size(lengths, num_buckets=3)
        >>> subsets = defined_split(data, split_func)
        >>> concat_dataset = ConcatDataset(subsets)
        >>> batch_sampler = BucketSampler(concat_dataset, 3 * [16])

    """

    def __init__(self, concat_dataset, batch_sizes, num_batches=None,
                 permuted_order=False, shuffle_each_bucket=True,
                 drop_last=False):
        self.datasets = _stack_concatenated_datasets(concat_dataset)
        self.batch_sizes = batch_sizes
        self.num_batches = num_batches
        self.permuted_order = permuted_order
        self.shuffle_each_bucket = shuffle_each_bucket
        self.drop_last = drop_last
        self.dataset_ids = list(range(len(self.datasets)))

        if isinstance(permuted_order, list):
            assert sorted(self.permuted_order) == self.dataset_ids, \
                '`permuted_order` not consistent with number of data sets.'
            self.dataset_ids = permuted_order

        assert all(
            [self.batch_sizes[i] > 0 and isinstance(self.batch_sizes[i], int)
             for i in self.dataset_ids]
        ), 'Only positive integers permitted for `num_batches`.'

        if self.num_batches is not None:
            assert all(
                [self.num_batches[i] >= 0 and isinstance(
                    self.num_batches[i], int) for i in self.dataset_ids
                 if self.num_batches[i] is not None]), \
                "Only non-negative integers or " \
                "`None` permitted for `num_batches`."

        if not isinstance(drop_last, bool):
            raise ValueError("drop_last should be a boolean value, but got "
                             "drop_last={}".format(drop_last))

    def __iter__(self):
        r"""Iterates sequentially over data sets and forms batches

        """
        all_batches = []
        batch = []

        if self.permuted_order is True:
            self.dataset_ids = list(np.random.permutation(self.dataset_ids))

        # iterate over data sets ordered by data set id
        for dset_id in self.dataset_ids:

            dataset = self.datasets[dset_id]
            num_batch = 0

            if self.shuffle_each_bucket:  # random samples from data set
                dataset = list(np.random.permutation(dataset))

            for sample in dataset:  # iterate over samples in data set
                if self.num_batches is not None and \
                        self.num_batches[dset_id] is not None:
                    if num_batch == self.num_batches[dset_id]:
                        break
                batch.append(sample)
                if len(batch) == self.batch_sizes[dset_id]:
                    all_batches.append(batch)
                    num_batch += 1
                    batch = []

            # yield full batch and also \
            # handle last samples of bucket which cannot form entire batch
            if len(batch) > 0 and not self.drop_last:
                all_batches.append(batch)
            batch = []

        return iter(all_batches)

    def __len__(self):

        sampler_size = 0
        for dset_id in self.dataset_ids:

            dataset = self.datasets[dset_id]
            bs = self.batch_sizes[dset_id]
            requested_batches = None if self.num_batches is None \
                else self.num_batches[dset_id]

            if self.drop_last:
                fitted_batches = len(dataset) // bs
            else:
                fitted_batches = (len(dataset) + bs - 1) // bs

            if requested_batches is None:
                sampler_size += fitted_batches
            else:
                sampler_size += fitted_batches if \
                    fitted_batches < requested_batches \
                    else requested_batches

        return sampler_size


def buckets_by_boundaries(key_values, bucket_boundaries):
    r"""Split samples into buckets based on key values using bucket boundaries.

    Note:
        A sample is sorted into bucket :math:`i` if for their key value
        holds:

            :math:`b_{i-1} <= \text{key value} < b_i`,
            where :math:`b_i` is `bucket boundary` at index :math:`i`

    Args:
        key_values (list): contains key values, e.g. sequence lengths
        bucket_boundaries (list): contains boundaries of buckets in
            ascending order. The list should neither contain a lower or
            upper boundary, e.g. not numpy.iinfo.min or numpy.iinfo.max.

    Returns:
        func: Key function to use for splitting: \
        :math:`f(\text{item}) = \text{bucket\_id}`

    Example:
        >>> lengths = [288, 258, 156, 99, 47, 13]
        >>> boundaries = [80, 150]
        >>> split_func = buckets_by_boundaries(lengths, boundaries)
        >>> [split_func(i) for i in range(len(lengths))]
        [2, 2, 2, 1, 0, 0]

    """
    assert bucket_boundaries == sorted(bucket_boundaries), \
        "Iterable `bucket_boundaries` not given in ascending order."

    assert len(bucket_boundaries) == \
        len(np.unique(bucket_boundaries)), \
        "Iterable `bucket_boundaries` contains duplicate(s)."

    num_buckets = len(bucket_boundaries) + 1

    def key_func(item):
        key_val = key_values[item]
        for bucket_id in range(num_buckets - 1):
            if key_val < bucket_boundaries[bucket_id]:
                return bucket_id
        return num_buckets - 1

    return key_func


def buckets_of_even_size(key_values, num_buckets, reverse=False):
    r"""Split samples into buckets of even size based on key values.

    The samples are sorted with either increasing (or decreasing) key value.
    If number of samples cannot be distributed evenly to buckets,
    the first buckets are filled up with one remainder each.

    Args:
        key_values (list): contains key values, e.g. sequence lengths
        num_buckets (int): number of buckets to form. Permitted are
            positive integers
        reverse (bool, optional): if True, then sort in descending order.
            Default: `False`

    Returns:
        func: Key function to use for splitting: \
        :math:`f(\text{item}) = \text{bucket\_id}`

    Example:
        >>> lengths = [288, 258, 156, 47, 112, 99, 13]
        >>> num_buckets = 4
        >>> split_func = buckets_of_even_size(lengths, num_buckets)
        >>> [split_func(i) for i in range(len(lengths))]
        [3, 2, 2, 0, 1, 1, 0]

    """
    # make sure that bucket size is larger than 0
    assert (len(key_values) >= num_buckets), \
        "Not enough `key_values` for `num_buckets` in order to" \
        " form even buckets."

    assert (num_buckets > 0) and isinstance(num_buckets, int), \
        "Specified value for `num_buckets` not a positive integer."

    bucket_size, remainder = divmod(len(key_values), num_buckets)
    sorted_indices = sorted(range(len(key_values)),
                            key=lambda i: key_values[i],
                            reverse=reverse)

    bucket_membership = len(key_values) * [0]
    sample_count, bucket_id, r = 0, 0, remainder

    for i in sorted_indices:
        bucket_membership[i] = bucket_id
        sample_count += 1
        if sample_count == bucket_size:  # bucket full
            # distribute one remainder per bucket
            if r > 0 and bucket_id + r == remainder:
                r -= 1
                sample_count -= 1
            else:
                sample_count = 0
                bucket_id += 1

    def key_func(item):
        return bucket_membership[item]

    return key_func


def _stack_concatenated_datasets(concat_dataset):
    r"""Extract and stack indices of different data sets from `concat_dataset`.

    Each data set is represented by a complete range of indices
    starting from 0...len(data set) for the first data set.
    The ranges of the following data sets build on
    the range of the corresponding previous data set so that
    the indices of the lists are cumulative::

        datasets = [[0 ... len(data_1) - 1],
                    [len(data_1) ... len(data_1) - 1], ... ]

    Args:
        concat_dataset (ConcatDataset): data set to sample from

    Returns:
        list: list of lists of data set indices

    Example:
        >>> import torch
        >>> from torch.utils.data import (TensorDataset, ConcatDataset)
        >>> data_1 = TensorDataset(torch.Tensor([0, 1, 2, 3]))
        >>> data_2 = TensorDataset(torch.Tensor([0, 1, 2]))
        >>> concat_dataset = ConcatDataset((data_1, data_2))
        >>> concat_dataset.cumulative_sizes
        [4, 7]
        >>> _stack_concatenated_datasets(concat_dataset)
        [[0, 1, 2, 3], [4, 5, 6]]

    """
    datasets = []
    start_idx = 0
    for upper_edge in concat_dataset.cumulative_sizes:
        datasets += [list(range(start_idx, upper_edge))]
        start_idx = upper_edge

    return datasets



import os
import subprocess
from warnings import warn
import urllib
import tarfile

from tqdm import tqdm
import numpy as np
import audiofile as af
from torch.utils.data import Subset

from ..utils import run_worker_threads


__doctest_skip__ = ['load']


def load(filename, *, duration=None, offset=0):
    r"""Load audio file.

    If an error occurrs during loading as the file could not be found,
    is empty, or has the wrong format an empty signal is returned and a warning
    shown.

    Args:
        file (str or int or file-like object): file name of input audio file
        duration (float, optional): return only a specified duration in
            seconds. Default: `None`
        offset (float, optional): start reading at offset in seconds.
            Default: `0`

    Returns:
        tuple:

            * **numpy.ndarray**: two-dimensional array with shape
              `(channels, samples)`
            * **int**: sample rate of the audio file

    Example:
        >>> signal, sampling_rate = load('speech.wav')

    """
    signal = np.array([[]])  # empty signal of shape (1, 0)
    sampling_rate = None
    try:
        signal, sampling_rate = af.read(filename,
                                        duration=duration,
                                        offset=offset,
                                        always_2d=True)
    except ValueError:
        warn('File opening error for: {}'.format(filename), UserWarning)
    except (IOError, FileNotFoundError):
        warn('File does not exist: {}'.format(filename), UserWarning)
    except RuntimeError:
        warn('Runtime error for file: {}'.format(filename), UserWarning)
    except subprocess.CalledProcessError:
        warn('ffmpeg conversion failed for: {}'.format(filename), UserWarning)
    return signal, sampling_rate


def download_url(url, root, *, filename=None, md5=None):
    r"""Download a file from an url to a specified directory.

    Args:
        url (str): URL to download file from
        root (str): directory to place downloaded file in
        filename (str, optional): name to save the file under.
            If `None`, use basename of URL. Default: `None`
        md5 (str, optional): MD5 checksum of the download.
            If None, do not check. Default: `None`

    Returns:
       str: path to downloaded file

    """
    root = safe_path(root)
    if not filename:
        filename = os.path.basename(url)
    filename = os.path.join(root, filename)

    os.makedirs(root, exist_ok=True)

    # downloads file
    if not os.path.isfile(filename):
        bar_updater = _gen_bar_updater(tqdm(unit='B', unit_scale=True))
        try:
            print('Downloading ' + url + ' to ' + filename)
            urllib.request.urlretrieve(url, filename, reporthook=bar_updater)
        except OSError:
            if url[:5] == 'https':
                url = url.replace('https:', 'http:')
                print('Failed download. Trying https -> http instead.'
                      ' Downloading ' + url + ' to ' + filename)
                urllib.request.urlretrieve(url, filename,
                                           reporthook=bar_updater)
    return safe_path(filename)


def download_url_list(urls, root, *, num_workers=0):
    r"""Download files from a list of URLs to a specified directory.

    Args:
        urls (list of str or dict): either list of URLs or dictionary
            with URLs as keys and with either filenames or tuples of
            filename and MD5 checksum as values. Uses basename of URL if
            filename is `None`. Performs no check if MD5 checksum is `None`
        root (str): directory to place downloaded files in
        num_workers (int, optional): number of worker threads
            (0 = len(urls)). Default: `0`

    """
    # always convert to dict
    if type(urls) is list:
        urls = {x: None for x in urls}

    # download file and extract
    def _task(url, filename):
        md5 = None
        if type(filename) is tuple:
            filename, md5 = filename
        return download_url(url, root, filename=filename, md5=md5)

    # start workers
    params = [(url, filename) for url, filename in urls.items()]
    return run_worker_threads(num_workers, _task, params)


def extract_archive(filename, *, out_path=None, remove_finished=False):
    r"""Extract archive.

    Currently `tar.gz` and `tar` archives are supported.

    Args:
        filename (str): path to archive
        out_path (str, optional): extract archive in this folder.
            Default: folder where archive is located in
        remove_finished (bool, optional): if `True` remove archive after
            extraction. Default: `False`

    """
    print('Extracting {}'.format(filename))
    if out_path is None:
        out_path = os.path.dirname(filename)
    if filename.endswith('tar.gz'):
        tar = tarfile.open(filename, 'r:gz')
    elif filename.endswith('tar'):
        tar = tarfile.open(filename, 'r:')
    else:
        raise RuntimeError('Archive format not supported.')
    tar.extractall(path=out_path)
    tar.close()
    if remove_finished:
        os.unlink(filename)


def sampling_rate_after_transform(dataset):
    r"""Sampling rate of data set after all transforms are applied.

    A change of sampling rate by a transform is only recognized, if that
    transform has the attribute :attr:`output_sampling_rate`.

    Args:
        dataset (torch.utils.data.Dataset): data set with `sampling_rate`
            attribute or property

    Returns:
        int: sampling rate in Hz after all transforms are applied

    Example:
        >>> from audtorch import datasets, transforms
        >>> t = transforms.Resample(input_sampling_rate=16000,
        ...                         output_sampling_rate=8000)
        >>> data = datasets.WhiteNoise(sampling_rate=16000, transform=t)
        >>> sampling_rate_after_transform(data)
        8000

    """
    sampling_rate = dataset.original_sampling_rate
    try:
        # List of composed transforms
        transforms = dataset.transform.transforms
    except AttributeError:
        # Single transform
        transforms = [dataset.transform]
    for transform in transforms:
        if hasattr(transform, 'output_sampling_rate'):
            sampling_rate = transform.output_sampling_rate
    return sampling_rate


def ensure_same_sampling_rate(datasets):
    r"""Raise error if provided data set differ in sampling rate.

    All data sets that are checked need to have a `sampling_rate` attribute or
    property.

    Args:
        datasets (list of torch.utils.data.Dataset): list of at least two audio
            data sets.

    """
    for dataset in datasets:
        if not hasattr(dataset, 'sampling_rate'):
            raise RuntimeError("{} doesn't have a `sampling_rate` attribute."
                               .format(dataset))
    for n in range(1, len(datasets)):
        if datasets[0].sampling_rate != datasets[n].sampling_rate:
            error_msg = 'Sampling rates do not match:\n'
            for dataset in datasets:
                info = dataset.__repr__()
                error_msg += '{}Hz from {}'.format(dataset.sampling_rate, info)
            raise ValueError(error_msg)


def ensure_df_columns_contain(df, labels):
    r"""Raise error if list of labels are not in dataframe columns.

    Args:
        df (pandas.dataframe): data frame
        labels (list of str): labels to be expected in `df.columns`

    Example:
        >>> import pandas as pd
        >>> df = pd.DataFrame(data=[(1, 2)], columns=['a', 'b'])
        >>> ensure_df_columns_contain(df, ['a', 'c'])
        Traceback (most recent call last):
        RuntimeError: Dataframe contains only these columns: 'a, b'

    """
    ensure_df_not_empty(df)
    if not set(labels) <= set(df.columns):
        raise RuntimeError("Dataframe contains only these columns: '{}'"
                           .format(', '.join(df.columns)))


def ensure_df_not_empty(df, labels=None):
    r"""Raise error if dataframe is empty.

    Args:
        df (pandas.dataframe): data frame
        labels (list of str, optional): list of labels used to shrink data
            set. Default: `None`

    Example:
        >>> import pandas as pd
        >>> df = pd.DataFrame()
        >>> ensure_df_not_empty(df)
        Traceback (most recent call last):
        RuntimeError: No valid data points found in data set

    """
    error_message = 'No valid data points found in data set'
    if labels is not None:
        error_message += (' for the selected labels: {}'
                          .format(', '.join(labels)))
    if len(df) == 0:
        raise RuntimeError(error_message)


def files_and_labels_from_df(df, *, root='.', column_labels='label',
                             column_filename='filename'):
    r"""Extract list of files and labels from dataframe columns.

    Args:
        df (pandas.DataFrame): data frame with filenames and labels. Relative
            from `root`
        root (str, optional): root directory of data set. Default: `.`
        column_labels (str or list of str, optional): name of data frame
            column(s) containing the desired labels. Default: `label`
        column_filename (str, optional): name of column holding the file
            names. Default: `filename`

    Returns:
        tuple:
            * list of str: list of files
            * list of str or list of dicts: list of labels

    Example:
        >>> import pandas as pd
        >>> df = pd.DataFrame(data=[('speech.wav', 'speech')],
        ...                   columns=['filename', 'label'])
        >>> files, labels = files_and_labels_from_df(df)
        >>> os.path.relpath(files[0]), labels[0]
        ('speech.wav', 'speech')

    """
    if df is None:
        return [], []
    root = safe_path(root)
    if isinstance(column_labels, str):
        column_labels = [column_labels]
    ensure_df_columns_contain(df, column_labels)
    df = df[[column_filename] + column_labels]
    # Drop empty entries
    df = df.dropna().reset_index(drop=True)
    ensure_df_not_empty(df, column_labels)
    # Assign files and labels
    files = df.pop(column_filename).tolist()
    files = [os.path.join(root, f) for f in files]
    if len(column_labels) == 1:
        # list of strings
        labels = df.values.T[0].tolist()
    else:
        # list of dicts
        labels = df.to_dict('records')
    return files, labels


def _gen_bar_updater(pbar):
    def bar_update(count, block_size, total_size):
        if pbar.total is None and total_size:
            pbar.total = total_size
        progress_bytes = count * block_size
        pbar.update(progress_bytes - pbar.n)

    return bar_update


def defined_split(dataset, split_func):
    r"""Split data set into desired non-overlapping subsets.

    Args:
        dataset (torch.utils.data.Dataset): data set to be split
        split_func (func): function mapping from data set index to subset id,
            :math:`f(\text{index}) = \text{subset\_id}`.
            The target domain of subset ids does not need to cover the
            complete range `[0, 1, ..., (num_subsets - 1)]`

    Returns:
        (list of Subsets): desired subsets according to :attr:`split_func`

    Example:
        >>> import torch
        >>> from torch.utils.data import TensorDataset
        >>> from audtorch.samplers import buckets_of_even_size
        >>> data = TensorDataset(torch.randn(100))
        >>> lengths = np.random.randint(0, 1000, (100,))
        >>> split_func = buckets_of_even_size(lengths, 5)
        >>> subsets = defined_split(data, split_func)
        >>> [len(subset) for subset in subsets]
        [20, 20, 20, 20, 20]

    """
    subset_ids = [split_func(i) for i in range(len(dataset))]
    unique_subset_ids = sorted(set(subset_ids))
    num_subsets = len(unique_subset_ids)

    split_indices = [[] for _ in range(num_subsets)]

    for i, subset_id in enumerate(subset_ids):
        # handle non-coherent target domain
        subset_id = unique_subset_ids.index(subset_id)
        split_indices[subset_id] += [i]

    return [Subset(dataset, indices)
            for indices in split_indices]


def safe_path(path):
    """Ensure the path is absolute and doesn't include `..` or `~`.

    Args:
        path (str): absolute or relative path

    Returns:
        str: absolute path

    """
    return os.path.abspath(os.path.expanduser(path))



import os

import pandas as pd

from audtorch.datasets.base import AudioDataset
from audtorch.datasets.utils import download_url


__doctest_skip__ = ['*']


class VoxCeleb1(AudioDataset):
    r"""VoxCeleb1 data set.

    Open and publicly available data set of voices from University of Oxford:
    http://www.robots.ox.ac.uk/~vgg/data/voxceleb/vox1.html

    VoxCeleb1 is a large audio-visual data set consisting of short clips of
    human speech extracted from YouTube interviews with celebrities. It is
    free for commercial and research purposes.

    Licence: CC BY-SA 4.0

    * :attr:`transform` controls the input transform
    * :attr:`target_transform` controls the target transform
    * :attr:`files` controls the audio files of the data set
    * :attr:`targets` controls the corresponding targets
    * :attr:`sampling_rate` holds the sampling rate of data set

    In addition, the following class attributes are available:

    * :attr:`url` holds its URL

    Args:
        root (str): root directory of dataset
        partition (str, optional): name of the data partition to use.
            Choose one of `train`, `dev`, `test` or `None`. If `None` is given,
            then the whole data set will be returned. Default: `train`
        transform (callable, optional): function/transform applied on the
            signal. Default: `None`
        target_transform (callable, optional): function/transform applied on
            the target. Default: `None`

    Note:
        * This data set will work only if the identification file is downloaded
          as is from the official homepage. Please open it in your browser and
          copy paste its contents in a file in your computer.
        * To download the data set go to
          http://www.robots.ox.ac.uk/~vgg/data/voxceleb/ and fill in the form
          to request a password. Get the Audio Files that the owners provide.

        * When using the VoxCeleb1 data set in your research, please cite
          the following publication: :cite:`nagrani2017voxceleb`.

    Example:
        >>> import sounddevice as sd
        >>> data = VoxCeleb1('/data/voxceleb1')
        >>> print(data)
        Dataset VoxCeleb1
            Number of data points: 138361
            Root Location: /data/voxceleb1
            Sampling Rate: 16000Hz
            Labels: speaker ID
        >>> signal, target = data[0]
        >>> target
        'id10003'
        >>> sd.play(signal.transpose(), data.sampling_rate)

    """
    url = ('http://www.robots.ox.ac.uk/~vgg/data/voxceleb/')
    _iden_file_url = (
        'http://www.robots.ox.ac.uk/~vgg/data/voxceleb/meta/iden_split.txt')
    _partitions = {'train': 1, 'dev': 2, 'test': 3}

    def __init__(self, root, *, partition='train',
                 transform=None, target_transform=None):
        super().__init__(root, files=[], targets=[], transform=transform,
                         sampling_rate=16000,
                         target_transform=target_transform)

        filelist = pd.read_csv(
            os.path.join(
                self.root,
                download_url(self._iden_file_url, self.root)),
            sep=' ', header=None)
        self.files, self.targets = self._get_files_speaker_lists(filelist)

        if partition is not None:
            # filter indices based on identification split
            indices = [index for index, x in enumerate(filelist[0])
                       if x == self._partitions[partition]]
            self.files = [self.files[index] for index in indices]
            self.targets = [self.targets[index] for index in indices]

    def _get_files_speaker_lists(self, filelist):
        r"""Extract file names and speaker IDs.

        Args:
            filelist (pandas.DataFrame): data frame containing file list
                and speakers

        Returns:
            list: files belonging to data set
            list: speaker IDs per file

        """
        files = [os.path.join(self.root, 'wav', x)
                 for x in filelist[1]]
        speakers = [x.split('/')[0] for x in filelist[1]]
        return files, speakers

    def extra_repr(self):
        fmt_str = '    Labels: speaker ID\n'
        return fmt_str



import pytest
import torch
from audtorch import collate


# data format FS = (feature dim, sequence dim)
batch_1 = [[torch.zeros(4, 5), torch.zeros(10)],
           [torch.zeros(4, 6), torch.zeros(12)]]
sequence_dimensions_1 = [-1, -1]
expected_1 = [[2, 4, 6], [2, 6, 4], [6, 2, 4]]

# data format CFS = (channel dim, feature dim, sequence dim)
batch_2 = [[torch.zeros(3, 4, 5), torch.zeros(10)],
           [torch.zeros(3, 4, 6), torch.zeros(12)]]
sequence_dimensions_2 = [2, 0]
expected_2 = [[2, 3, 4, 6], [2, 6, 3, 4], [6, 2, 3, 4]]


@pytest.mark.parametrize("batch,sequence_dimensions,expected",
                         [(batch_1, sequence_dimensions_1, expected_1),
                          (batch_2, sequence_dimensions_2, expected_2)])
@pytest.mark.parametrize("batch_first", [None, True, False])
def test_seq2seq(batch, sequence_dimensions, expected, batch_first):

    collation = collate.Seq2Seq(
        sequence_dimensions=sequence_dimensions,
        batch_first=batch_first,
        pad_values=[-1, -1])
    output = collation(batch)

    if batch_first is None:
        assert list(output[0].shape) == expected[0]

    if batch_first:
        assert list(output[0].shape) == expected[1]

    if batch_first is False:
        assert list(output[0].shape) == expected[2]



import pytest
import numpy as np
import pandas as pd
import torch
from torch.utils.data import TensorDataset

from audtorch import (datasets, samplers, transforms)


xfail = pytest.mark.xfail
filterwarnings = pytest.mark.filterwarnings


# --- datasets/noise.py ---
@pytest.mark.parametrize('duration', [0.01, 0.1, 1])
@pytest.mark.parametrize('sampling_rate', [8000])
@pytest.mark.parametrize('mean', [0, 1])
@pytest.mark.parametrize('stdev', [1, 0.5])
def test_whitenoise(duration, sampling_rate, mean, stdev):
    dataset = datasets.WhiteNoise(duration=duration,
                                  sampling_rate=sampling_rate,
                                  mean=mean,
                                  stdev=stdev)
    noise, label = next(iter(dataset))
    samples = int(np.ceil(duration * sampling_rate))
    assert noise.shape == (1, samples)
    assert label == 'white noise'
    assert -1 <= np.max(np.abs(noise)) <= 1
    assert len(dataset) == 1


# --- datasets/utils.py ---
crop = transforms.RandomCrop(8192)
resamp1 = transforms.Resample(48000, 44100)
resamp2 = transforms.Resample(44100, 16000)
t1 = transforms.Compose([crop, resamp1])
t2 = transforms.Compose([crop, resamp1, resamp2])
t3 = transforms.Compose([resamp1, crop, resamp2])
d0 = datasets.WhiteNoise(duration=0.5, sampling_rate=48000, transform=crop)
d1 = datasets.WhiteNoise(duration=0.5, sampling_rate=48000, transform=t1)
d2 = datasets.WhiteNoise(duration=0.5, sampling_rate=48000, transform=t2)
d3 = datasets.WhiteNoise(duration=0.5, sampling_rate=48000, transform=t3)
df_empty = pd.DataFrame()
df_a = pd.DataFrame(data=[0], columns=['a'])
df_ab = pd.DataFrame(data=[('0', 1)], columns=['a', 'b'])


@pytest.mark.parametrize('list_of_datasets', [
    (d2, d3),
    pytest.param([d0, d1], marks=xfail(raises=ValueError))
])
def test_audioconcatdataset(list_of_datasets):
    datasets.AudioConcatDataset(list_of_datasets)


@pytest.mark.parametrize(('input,expected_output,expected_sampling_rate,'
                          'expected_warning'), [
    ('', np.array([[]]), None, 'File does not exist: '),
])
def test_load(input, expected_output, expected_sampling_rate,
              expected_warning):
    with pytest.warns(UserWarning, match=expected_warning):
        output, sampling_rate = datasets.load(input)
    assert np.array_equal(output, expected_output)
    assert sampling_rate == expected_sampling_rate


@pytest.mark.parametrize('transform', [t1, t2, t3])
def test_sampling_rate_after_transform(transform):
    expected_sampling_rate = transform.transforms[-1].output_sampling_rate
    dataset = datasets.WhiteNoise(duration=0.5,
                                  sampling_rate=48000,
                                  transform=transform)
    sampling_rate = datasets.sampling_rate_after_transform(dataset)
    assert sampling_rate == expected_sampling_rate
    assert dataset.sampling_rate == expected_sampling_rate


@pytest.mark.parametrize('list_of_datasets', [
    [d0, d0, d0],
    [d1, d1],
    [d2, d2],
    pytest.param([1], marks=xfail(raises=RuntimeError)),
    pytest.param([d0, d1], marks=xfail(raises=ValueError)),
    pytest.param([d1, d2], marks=xfail(raises=ValueError)),
])
def test_ensure_same_sampling_rate(list_of_datasets):
    datasets.ensure_same_sampling_rate(list_of_datasets)


@pytest.mark.parametrize('df,labels', [
    pytest.param(df_empty, 'a', marks=xfail(raises=RuntimeError)),
    (df_a, 'a'),
    pytest.param(df_a, 'b', marks=xfail(raises=RuntimeError)),
    (df_ab, ['a', 'b']),
    (df_ab, 'a'),
    pytest.param(df_ab, 'c', marks=xfail(raises=RuntimeError)),
])
def test_ensure_df_columns_contain(df, labels):
    datasets.ensure_df_columns_contain(df, labels)


@pytest.mark.parametrize('df', [
    pytest.param(df_empty, marks=xfail(raises=RuntimeError)),
    df_a,
])
def test_ensure_df_not_empty(df):
    datasets.ensure_df_not_empty(df)


@pytest.mark.parametrize('df,expected_files,expected_labels', [
    (df_ab, ['./0'], [1]),
    (None, [], []),
    pytest.param(df_empty, [], [], marks=xfail(raises=RuntimeError)),
])
def test_files_and_labels_from_df(df, expected_files, expected_labels):
    files, labels = datasets.files_and_labels_from_df(df,
                                                      root='.',
                                                      column_filename='a',
                                                      column_labels='b')
    expected_files = [datasets.safe_path(f) for f in expected_files]
    assert files == expected_files
    assert labels == expected_labels


@pytest.mark.parametrize('key_values,split_func,kwargs,expected', [
    (list(range(5)), samplers.buckets_of_even_size, 2, [3, 2]),
    (list(range(10)), samplers.buckets_by_boundaries, [0, 3, 8], [3, 5, 2])
])
def test_defined_split(key_values, split_func, kwargs, expected):
    data = TensorDataset(torch.arange(len(key_values)))
    split_func = split_func(torch.Tensor(key_values), kwargs)
    subsets = datasets.defined_split(data, split_func)

    assert expected == [len(subset) for subset in subsets]



import pytest
import numpy as np
import scipy
import resampy
import librosa

import audtorch.transforms as transforms
import audtorch.transforms.functional as F


xfail = pytest.mark.xfail

a11 = np.array([1, 2, 3, 4], dtype=float)
a12 = np.array([5, 6, 7, 8], dtype=float)
a21 = np.array([9, 10, 11, 12], dtype=float)
a22 = np.array([13, 14, 15, 16], dtype=float)
ones = np.ones(4)
zeros = np.zeros(4)
A = np.array([[a11, a12], [a21, a22]])  # Tensor of shape (2, 2, 4)

# Ratio in dB to add two signals to yield 1.5 magnitude
_half_ratio = -10 * np.log10(0.5 ** 2)


def _mean(signal, axis):
    """Return mean along axis and preserve number of dimensions."""
    return np.expand_dims(np.mean(signal, axis=axis), axis=axis)


def _pad(vector, padding, value):
    """Add padding to a vector using np.pad."""
    return np.pad(vector, padding, 'constant', constant_values=value)


@pytest.mark.parametrize('input,idx,axis,expected_output', [
    (A, (0, 2), -1, np.array([[a11[:2], a12[:2]], [a21[:2], a22[:2]]])),
    (A, (1, 2), -2, np.array([[a12], [a22]])),
    (A, (0, 1), 0, np.array([[a11, a12]])),
    (A, -1, -2, np.array([[a12], [a22]])),
    (A, 0, -2, np.array([[a11], [a21]])),
    (a11, (1, 2), -1, a11[1:2]),
])
def test_crop(input, idx, axis, expected_output):
    t = transforms.Crop(idx, axis=axis)
    assert np.array_equal(t(input), expected_output)


@pytest.mark.parametrize('input,padding,value,axis,expected_output', [
    (A, 1, 0, -1, np.array([[_pad(a11, 1, 0), _pad(a12, 1, 0)],
                            [_pad(a21, 1, 0), _pad(a22, 1, 0)]])),
    (A, (0, 1), 1, 1, np.array([[a11, a12, ones], [a21, a22, ones]])),
    (A, (1, 0), 0, 0, np.array([[zeros, zeros], [a11, a12], [a21, a22]])),
    (a11, 1, 1, -1, _pad(a11, 1, 1)),
])
def test_pad(input, padding, value, axis, expected_output):
    t = transforms.Pad(padding, value=value, axis=axis)
    assert np.array_equal(t(input), expected_output)


@pytest.mark.parametrize('input,repetitions,axis', [
    (A, 2, -1),
    (A, 3, 0),
    (A, 4, 1),
    (a11, 1, 0)])
def test_replicate(input, repetitions, axis):
    expected_output = np.concatenate(tuple([input] * repetitions), axis)
    t = transforms.Replicate(repetitions, axis=axis)
    assert np.array_equal(t(input), expected_output)


@pytest.mark.parametrize('input,size,axis,method', [
    (A, 12, 0, 'pad'),
    (A, 12, 1, 'pad'),
    (A, 12, 2, 'pad'),
    (A, 12, 0, 'replicate'),
    (A, 12, 1, 'replicate'),
    (A, 12, 2, 'replicate')])
def test_expand(input, size, axis, method):
    t = transforms.Expand(size=size, axis=axis, method=method)
    if method == 'pad':
        assert np.array_equal(t(input), F.pad(
            input, (0, size - input.shape[axis]), axis=axis))
    else:
        assert np.array_equal(t(input), F.crop(F.replicate(
            input,
            repetitions=size // input.shape[axis] + 1,
            axis=axis), (0, size), axis=axis))


@pytest.mark.parametrize('input,channels,method,axis,expected_output', [
    (A, 2, 'mean', -2, A),
    (A, 1, 'mean', -2, _mean(A, axis=-2)),
    (A, 1, 'crop', -2, np.array([[a11], [a21]])),
    (A, 0, 'mean', -2, np.empty((2, 0, 4))),  # empty array with correct shape
    (A, 0, 'crop', -2, np.empty((2, 0, 4))),
    (A, 1, 'mean', 0, _mean(A, axis=0)),
    (a11, 1, 'crop', -1, np.array([a11[0]])),
    (a11, 1, 'crop', -2, np.array(a11)),
])
def test_downmix(input, channels, method, axis, expected_output):
    t = transforms.Downmix(channels, method=method, axis=axis)
    assert np.array_equal(t(input), expected_output)


@pytest.mark.parametrize('input,channels,method,axis,expected_output', [
    (A, 2, 'mean', -2, A),
    (A, 3, 'mean', -2, np.hstack((A, _mean(A, axis=-2)))),
    (A, 3, 'zero', -2, np.hstack((A, [[zeros], [zeros]]))),
    (A, 3, 'repeat', -2, np.hstack((A, A[:, -1, None, :]))),
    (A, 3, 'mean', 0, np.vstack((A, _mean(A, axis=0)))),
    (a11, 2, 'repeat', -2, np.array([a11, a11])),
])
def test_upmix(input, channels, method, axis, expected_output):
    t = transforms.Upmix(channels, method=method, axis=axis)
    assert np.array_equal(t(input), expected_output)


@pytest.mark.parametrize('input,channels,axis,expected_output', [
    (A, 2, -2, A),
    (A, 3, -2, np.hstack((A, _mean(A, axis=-2)))),
    (A, 3, 0, np.vstack((A, _mean(A, axis=0)))),
    (A, 1, -2, _mean(A, axis=-2)),
    (A, 0, -2, np.empty((2, 0, 4))),  # empty array with correct shape
    (A, 1, 0, _mean(A, axis=0)),
])
def test_remix(input, channels, axis, expected_output):
    t = transforms.Remix(channels, axis=axis)
    assert np.array_equal(t(input), expected_output)


@pytest.mark.parametrize('input,axis,expected_output', [
    (A, None, A / np.max(A)),
    (A, -1, np.array([[a11 / max(a11), a12 / max(a12)],
                      [a21 / max(a21), a22 / max(a22)]])),
    (a11, None, a11 / np.max(a11)),
    (a11, -1, a11 / np.max(a11)),
])
def test_normalize(input, axis, expected_output):
    t = transforms.Normalize(axis=axis)
    assert np.array_equal(t(input), expected_output)


@pytest.mark.parametrize('input,axis,mean,std', [
    (A, None, True, True),
    (A, -1, True, True),
    (a11, None, True, True),
    (a11, -1, True, True),
    (A, -1, False, True),
    (A, -1, True, False),
    (A, -1, False, False),
])
def test_standardize(input, axis, mean, std):
    t = transforms.Standardize(axis=axis)
    output = t(input)
    if mean:
        np.testing.assert_almost_equal(output.mean(axis=axis).mean(), 0)
    if std:
        np.testing.assert_almost_equal(output.std(axis=axis).mean(), 1)


@pytest.mark.parametrize('input,idx,axis', [
    (A, (0, 2), -1),
    (a11, (1, 2), -1),
])
def test_compose(input, idx, axis):
    t = transforms.Compose([transforms.Crop(idx, axis=axis),
                            transforms.Normalize(axis=axis)])
    expected_output = F.crop(input, idx, axis=axis)
    expected_output = F.normalize(expected_output, axis=axis)
    assert np.array_equal(t(input), expected_output)


@pytest.mark.parametrize('input,size,axis', [
    (A, 2, -1),
    (A, 1, -2),
    (A, 1, 0),
    (A, 0, -2),
    (a11, 3, -1),
])
def test_randomcrop(input, size, axis):
    t = transforms.RandomCrop(size, axis=axis)
    t.fix_randomization = True
    assert np.array_equal(t(input), t(input))
    assert np.array_equal(t(input), F.crop(input, t.idx, axis=t.axis))


@pytest.mark.parametrize('input,padding,value,axis', [
    (A, 1, 0, -1),
    (A, 2, 1, 1),
    (A, 0, 0, 0),
    (a11, 1, 1, -1),
])
def test_randompad(input, padding, value, axis):
    t = transforms.RandomPad(padding, value=value, axis=axis)
    t.fix_randomization = True
    assert np.array_equal(t(input), t(input))
    expected_output = F.pad(input, t.pad, value=t.value, axis=t.axis)
    assert np.array_equal(t(input), expected_output)


@pytest.mark.parametrize('input,input_sample_rate,output_sample_rate,axis', [
    (A, 4, 2, -1),
    (np.ones([4, 4, 2]), 4, 2, -2),
    (a11, 3, 2, -1),
])
@pytest.mark.parametrize('method', ['kaiser_best', 'kaiser_fast', 'scipy'])
def test_resample(input, input_sample_rate, output_sample_rate, method, axis):
    t = transforms.Resample(input_sample_rate, output_sample_rate,
                            method=method, axis=axis)
    output_length = int(input.shape[axis] * output_sample_rate
                        / float(input_sample_rate))
    print(input.shape)
    if method == 'scipy':
        expected_output = scipy.signal.resample(input, output_length,
                                                axis=axis)
    else:
        expected_output = resampy.resample(input, input_sample_rate,
                                           output_sample_rate, method=method,
                                           axis=axis)
    transformed_input = t(input)
    assert transformed_input.shape[axis] == output_length
    assert np.array_equal(transformed_input, expected_output)


@pytest.mark.parametrize('input,window_size,hop_size,axis', [
    (A, 4, 1, 2),
    pytest.param(A, 2048, 1024, 2, marks=xfail(raises=ValueError)),
    (np.random.normal(size=[2, 3, 16000]), 2048, 1024, 2),
    (np.random.normal(size=[2, 16000, 3]), 2048, 1024, 1),
    (np.random.normal(size=[16000, 2, 3]), 2048, 1024, 0),
    (np.random.normal(size=[16000, 2, 3]), 2048, 1024, 0),
    (np.random.normal(size=16000), 2048, 1024, 0),
])
def test_stft(input, window_size, hop_size, axis):
    t = transforms.Spectrogram(window_size, hop_size, axis=axis)
    spectrogram = F.stft(input, window_size, hop_size, axis=axis)
    magnitude, phase = librosa.magphase(spectrogram)
    assert np.array_equal(t(input), magnitude)
    assert np.array_equal(t.phase, phase)


@pytest.mark.parametrize('input,magnitude_boost', [
    (A, 1e-07),
    pytest.param(A, -1e-07, marks=xfail(raises=ValueError)),
    pytest.param(A, -1, marks=xfail(raises=ValueError)),
    (np.random.normal(size=[2, 3, 16000]), 1e-03),
    (np.random.normal(size=[2, 16000, 3]), 1e-07),
    (np.random.normal(size=[16000, 2, 3]), 1e-07),
    (np.random.normal(size=16000), 1e-07),
])
def test_log(input, magnitude_boost):
    input = input + abs(input.min())
    t_log = transforms.Log(magnitude_boost=magnitude_boost)
    assert np.array_equal(t_log(input), np.log(input + magnitude_boost))


@pytest.mark.parametrize('signal1,signal2,ratio,'
                         'percentage_silence,expected_signal', [
                             (A, A, 0, 0, 2 * A),
                             (A, A, _half_ratio, 0, 1.5 * A),
                             (a11, a11, 0, 0, 2 * a11),
                             (a11, a11, _half_ratio, 0, 1.5 * a11),
                             (A, np.zeros_like(A), 0, 0, A),
                             (np.zeros_like(A), A, 0, 0, A),
                             (A, np.zeros_like(A), 0, 1, A),
                         ])
def test_randomadditivemix(signal1, signal2, ratio, percentage_silence,
                           expected_signal):
    mixed_signal = F.additive_mix(signal1, signal2, ratio)
    assert np.array_equal(mixed_signal, expected_signal)
    augmentation_data_set = [[signal2]]
    t = transforms.RandomAdditiveMix(dataset=augmentation_data_set,
                                     ratios=[ratio],
                                     percentage_silence=percentage_silence)
    assert np.array_equal(t(signal1), expected_signal)
    t = transforms.RandomAdditiveMix(dataset=augmentation_data_set)
    transform = t(signal1)
    functional = F.additive_mix(signal1, signal2, t.ratio)
    assert np.array_equal(transform, functional)



import os

from .utils import (download_url, extract_archive, safe_path)
from .base import CsvDataset


__doctest_skip__ = ['*']


class MozillaCommonVoice(CsvDataset):
    """Mozilla Common Voice speech data set.

    Open and publicly available data set of voices from Mozilla:
    https://voice.mozilla.org/en/datasets

    License: CC-0 (public domain)

    Mozilla Common Voice includes the labels `text`, `up_votes`,
    `down_votes`, `age`, `gender`, `accent`, `duration`. You can select one of
    those labels which is returned as a string by the data set as target or you
    can specify a list of the labels and the data set will return a dictionary
    containing those labels. The default label that is returned is `text`.

    * :attr:`root` holds the data set's location
    * :attr:`transform` controls the input transform
    * :attr:`target_transform` controls the target transform
    * :attr:`files` controls the audio files of the data set
    * :attr:`targets` controls the corresponding targets
    * :attr:`sampling_rate` holds the sampling rate of the returned data
    * :attr:`original_sampling_rate` holds the sampling rate of the audio files
      of the data set

    In addition, the following class attribute is available

    * :attr:`url` holds the download link of the data set

    Args:
        root (str): root directory of data set, where the CSV files are
            located, e.g. `/data/MozillaCommonVoice/cv_corpus_v1`
        csv_file (str, optional): name of a CSV file from the `root`
            folder. No absolute path is possible. You are most probably
            interested in `cv-valid-train.csv`, `cv-valid-dev.csv`, and
            `cv-valid-test.csv`. Default: `cv-valid-train.csv`.
        label_type (str or list of str, optional): one of `text`, `up_votes`,
            `down_votes`, `age`, `gender`, `accent`, `duration`. Or a list of
            any combination of those. Default: `text`
        transform (callable, optional): function/transform applied on the
            signal. Default: `None`
        target_transform (callable, optional): function/transform applied on
            the target. Default: `None`
        download (bool, optional): download data set if not present.
            Default: `False`

    Note:
        The Mozilla Common Voice data set is constantly growing. If you
        choose to download it, it will always grep the latest version. If
        you require reproducibility of your results, make sure to store a
        safe snapshot of the version you used.

    Example:
        >>> import sounddevice as sd
        >>> data = MozillaCommonVoice('/data/MozillaCommonVoice/cv_corpus_v1')
        >>> print(data)
        Dataset MozillaCommonVoice
            Number of data points: 195776
            Root Location: /data/MozillaCommonVoice/cv_corpus_v1
            Sampling Rate: 48000Hz
            Labels: text
            CSV file: cv-valid-train.csv
        >>> signal, target = data[0]
        >>> target
        'learn to recognize omens and follow them the old king had said'
        >>> sd.play(signal.transpose(), data.sampling_rate)

    """  # noqa: E501

    url = ('https://common-voice-data-download.s3.amazonaws.com/'
           'cv_corpus_v1.tar.gz')

    def __init__(self, root, *, csv_file='cv-valid-train.csv',
                 label_type='text', transform=None, target_transform=None,
                 download=False):

        if download:
            self.root = safe_path(root)
            self._download()

        super().__init__(root, csv_file, sampling_rate=48000, sep=',',
                         column_labels=label_type, column_filename='filename',
                         transform=transform,
                         target_transform=target_transform)

    def _download(self):
        if self._check_exists():
            return
        download_dir = self.root
        corpus = 'cv_corpus_v1'
        if download_dir.endswith(corpus):
            download_dir = download_dir[:-len(corpus)]
        filename = download_url(self.url, download_dir)
        extract_archive(filename, remove_finished=True)



def pearsonr(x, y, batch_first=True):
    r"""Computes Pearson Correlation Coefficient across rows.

    Pearson Correlation Coefficient (also known as Linear Correlation
    Coefficient or Pearson's :math:`\rho`) is computed as:

    .. math::

        \rho = \frac {E[(X-\mu_X)(Y-\mu_Y)]} {\sigma_X\sigma_Y}

    If inputs are matrices, then then we assume that we are given a
    mini-batch of sequences, and the correlation coefficient is
    computed for each sequence independently and returned as a vector. If
    `batch_fist` is `True`, then we assume that every row represents a
    sequence in the mini-batch, otherwise we assume that batch information
    is in the columns.

    Warning:
        We do not account for the multi-dimensional case. This function has
        been tested only for the 2D case, either in `batch_first==True` or in
        `batch_first==False` mode. In the multi-dimensional case,
        it is possible that the values returned will be meaningless.

    Args:
        x (torch.Tensor): input tensor
        y (torch.Tensor): target tensor
        batch_first (bool, optional): controls if batch dimension is first.
            Default: `True`

    Returns:
        torch.Tensor: correlation coefficient between `x` and `y`

    Note:
        :math:`\sigma_X` is computed using **PyTorch** builtin
        **Tensor.std()**, which by default uses Bessel correction:

        .. math::

            \sigma_X=\displaystyle\frac{1}{N-1}\sum_{i=1}^N({x_i}-\bar{x})^2

        We therefore account for this correction in the computation of the
        covariance by multiplying it with :math:`\frac{1}{N-1}`.

    Shape:
        - Input: :math:`(N, M)` for correlation between matrices,
          or :math:`(M)` for correlation between vectors
        - Target: :math:`(N, M)` or :math:`(M)`. Must be identical to input
        - Output: :math:`(N, 1)` for correlation between matrices,
          or :math:`(1)` for correlation between vectors

    Examples:
        >>> import torch
        >>> _ = torch.manual_seed(0)
        >>> input = torch.rand(3, 5)
        >>> target = torch.rand(3, 5)
        >>> output = pearsonr(input, target)
        >>> print('Pearson Correlation between input and target is {0}'.format(output[:, 0]))
        Pearson Correlation between input and target is tensor([ 0.2991, -0.8471,  0.9138])

    """  # noqa: E501
    assert x.shape == y.shape

    if batch_first:
        dim = -1
    else:
        dim = 0

    centered_x = x - x.mean(dim=dim, keepdim=True)
    centered_y = y - y.mean(dim=dim, keepdim=True)

    covariance = (centered_x * centered_y).sum(dim=dim, keepdim=True)

    bessel_corrected_covariance = covariance / (x.shape[dim] - 1)

    x_std = x.std(dim=dim, keepdim=True)
    y_std = y.std(dim=dim, keepdim=True)

    corr = bessel_corrected_covariance / (x_std * y_std)

    return corr


def concordance_cc(x, y, batch_first=True):
    r"""Computes Concordance Correlation Coefficient across rows.

    Concordance Correlation Coefficient is computed as:

    .. math::

        \rho_c = \frac {2\rho\sigma_X\sigma_Y} {\sigma_X\sigma_X +
        \sigma_Y\sigma_Y + (\mu_X - \mu_Y)^2}

    where :math:`\rho` is Pearson Correlation Coefficient, :math:`\sigma_X`,
    :math:`\sigma_Y` are the standard deviation, and :math:`\mu_X`,
    :math:`\mu_Y` the mean values of :math:`X` and :math:`Y` accordingly.

    If inputs are matrices, then then we assume that we are given a
    mini-batch of sequences, and the concordance correlation coefficient is
    computed for each sequence independently and returned as a vector. If
    `batch_fist` is `True`, then we assume that every row represents a
    sequence in the mini-batch, otherwise we assume that batch information
    is in the columns.

    Warning:
        We do not account for the multi-dimensional case. This function has
        been tested only for the 2D case, either in `batch_first==True` or in
        `batch_first==False` mode. In the multi-dimensional case,
        it is possible that the values returned will be meaningless.

    Note:
        :math:`\sigma_X` is computed using **PyTorch** builtin
        **Tensor.std()**, which by default uses Bessel correction:

        .. math::

            \sigma_X=\displaystyle\frac{1}{N-1}\sum_{i=1}^N({x_i}-\bar{x})^2

        We therefore account for this correction in the computation of the
        concordance correlation coefficient by multiplying all standard
        deviations with :math:`\frac{N-1}{N}`. This is equivalent to
        multiplying only :math:`(\mu_X - \mu_Y)^2` with :math:`\frac{N}{
        N-1}`. We choose that option for numerical stability.

    Args:
        x (torch.Tensor): input tensor
        y (torch.Tensor): target tensor
        batch_first (bool, optional): controls if batch dimension is first.
            Default: `True`

    Returns:
        torch.Tensor: concordance correlation coefficient between `x` and `y`

    Shape:
        - Input: :math:`(N, M)` for correlation between matrices,
          or :math:`(M)` for correlation between vectors
        - Target: :math:`(N, M)` or :math:`(M)`. Must be identical to input
        - Output: :math:`(N, 1)` for correlation between matrices,
          or :math:`(1)` for correlation between vectors

    Examples:
        >>> import torch
        >>> _ = torch.manual_seed(0)
        >>> input = torch.rand(3, 5)
        >>> target = torch.rand(3, 5)
        >>> output = concordance_cc(input, target)
        >>> print('Concordance Correlation between input and target is {0}'.format(output[:, 0]))
        Concordance Correlation between input and target is tensor([ 0.2605, -0.7862,  0.5298])

    """  # noqa: E501
    assert x.shape == y.shape

    if batch_first:
        dim = -1
    else:
        dim = 0

    bessel_correction_term = (x.shape[dim] - 1) / x.shape[dim]

    r = pearsonr(x, y, batch_first)
    x_mean = x.mean(dim=dim, keepdim=True)
    y_mean = y.mean(dim=dim, keepdim=True)
    x_std = x.std(dim=dim, keepdim=True)
    y_std = y.std(dim=dim, keepdim=True)
    ccc = 2 * r * x_std * y_std / (x_std * x_std
                                   + y_std * y_std
                                   + (x_mean - y_mean)
                                   * (x_mean - y_mean)
                                   / bessel_correction_term)
    return ccc



import pytest
import scipy
import torch

import numpy as np

import audtorch.metrics as metrics
import audtorch.metrics.functional as F


@pytest.mark.parametrize('reduction', ['none', 'sum', 'mean'])
def test_energypreservingloss(reduction):
    loss = metrics.EnergyConservingLoss(reduction)
    # Random integers as tensors to avoid precision problems with torch.equal
    input = torch.rand((3, 5), requires_grad=True)
    target = torch.rand((3, 5))
    mixture = torch.rand((3, 5))
    noise = mixture - target
    noise_predicted = mixture - input
    expected_output = (torch.abs(input - target)
                       + torch.abs(noise - noise_predicted))
    output = loss(input, target, mixture)
    if reduction == 'none':
        assert torch.equal(output, expected_output)
    elif reduction == 'sum':
        assert torch.equal(output, torch.sum(expected_output))
    elif reduction == 'mean':
        assert torch.equal(output, torch.mean(expected_output))


@pytest.mark.parametrize('shape', [(5,), (5, 3)])
def test_pearsonr(shape):
    input = torch.rand(shape)
    target = torch.rand(shape)

    if len(shape) == 1:
        r = F.pearsonr(input, target)
        assert r.shape[0] == 1
        np.testing.assert_almost_equal(
            r.numpy()[0], scipy.stats.pearsonr(
                input.numpy(), target.numpy())[0], decimal=6)
    else:
        r = F.pearsonr(input, target)
        assert r.shape[0] == shape[0]
        for index, (input_row, target_row) in enumerate(zip(input, target)):
            np.testing.assert_almost_equal(
                r[index].numpy()[0], scipy.stats.pearsonr(
                    input_row.numpy(), target_row.numpy())[0], decimal=6)

        r = F.pearsonr(input, target, batch_first=False)
        assert r.shape[1] == shape[1]
        for index, (input_col, target_col) in enumerate(
                zip(input.transpose(0, 1), target.transpose(0, 1))):
            np.testing.assert_almost_equal(
                r[:, index].numpy()[0], scipy.stats.pearsonr(
                    input_col.numpy(), target_col.numpy())[0], decimal=6)


@pytest.mark.parametrize('shape', [(5,), (5, 3)])
def test_concordance_cc(shape):
    input = torch.rand(shape)
    target = torch.rand(shape)

    def concordance_cc(x, y):
        r = scipy.stats.pearsonr(x, y)[0]
        ccc = 2 * r * x.std() * y.std() / (x.std() * x.std()
                                           + y.std() * y.std()
                                           + (x.mean() - y.mean())
                                           * (x.mean() - y.mean()))
        return ccc

    if len(shape) == 1:
        ccc = F.concordance_cc(input, target)
        assert ccc.shape[0] == 1
        np.testing.assert_almost_equal(
            ccc.numpy()[0], concordance_cc(input.numpy(), target.numpy()),
            decimal=6)
    else:
        ccc = F.concordance_cc(input, target)
        assert ccc.shape[0] == shape[0]
        for index, (input_row, target_row) in enumerate(zip(input, target)):
            np.testing.assert_almost_equal(
                ccc[index].numpy()[0], concordance_cc(
                    input_row.numpy(), target_row.numpy()), decimal=6)

        ccc = F.concordance_cc(input, target, batch_first=False)
        assert ccc.shape[1] == shape[1]
        for index, (input_col, target_col) in enumerate(
                zip(input.transpose(0, 1), target.transpose(0, 1))):
            np.testing.assert_almost_equal(
                ccc[:, index].numpy()[0], concordance_cc(
                    input_col.numpy(), target_col.numpy()), decimal=6)



from setuptools import setup, find_packages

setup(
    name='audtorch',
    packages=find_packages(),
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    install_requires=[
        'numpy',
        'audiofile',
        'resampy',
        'torch',
        'pandas',
        'tqdm',
        'tabulate',
    ],
    author=('Andreas Triantafyllopoulos, '
            'Stephan Huber, '
            'Johannes Wagner, '
            'Hagen Wierstorf'),
    author_email='atriant@audeering.com',
    description='Deep learning with PyTorch and audio',
    long_description=open('README.rst').read(),
    long_description_content_type='text/x-rst',
    license='MIT License',
    keywords=['audio'],
    url='https://github.com/audeering/audtorch',
    project_urls={
        'Documentation': 'https://audtorch.readthedocs.io',
        'Tracker': 'https://github.com/audeering/audtorch/issues',
    },
    platforms='any',
    tests_require=['pytest'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Multimedia :: Sound/Audio',
    ],
)



import random
from warnings import warn

import numpy as np
import resampy
try:
    import librosa
except ImportError:
    librosa = None
try:
    import scipy
except ImportError:
    scipy = None

from . import functional as F


class Compose(object):
    r"""Compose several transforms together.

    Args:
        transforms (list of object): list of transforms to compose
        fix_randomization (bool, optional): controls randomization of
            underlying transforms. Default: `False`

    Example:
        >>> a = np.array([[1, 2], [3, 4]])
        >>> t = Compose([Crop(-1), Pad(1)])
        >>> print(t)
        Compose(
            Crop(idx=-1, axis=-1)
            Pad(padding=1, value=0, axis=-1)
        )
        >>> t(a)
        array([[0, 2, 0],
               [0, 4, 0]])

    """

    def __init__(self, transforms, *, fix_randomization=False):
        self.transforms = transforms
        self.fix_randomization = fix_randomization

    def __call__(self, signal):
        for t in self.transforms:
            if hasattr(t, 'fix_randomization'):
                t.fix_randomization = self.fix_randomization
            signal = t(signal)
        return signal

    def __repr__(self):
        format_string = self.__class__.__name__ + '('
        for t in self.transforms:
            format_string += '\n'
            format_string += '    {0}'.format(t)
        format_string += '\n)'
        if self.fix_randomization:
            format_string += ', fix_randomization=True'
        return format_string


class Crop(object):
    r"""Crop along an axis.

    * :attr:`idx` controls the index for cropping
    * :attr:`axis` controls axis of cropping

    Args:
        idx (int or tuple): first (and last) index to return
        axis (int, optional): axis along to crop. Default: `-1`

    Note:
        Indexing from the end with `-1`, `-2`, ... is allowed. But you cannot
        use `-1` in the second part of the tuple to specify the last entry.
        Instead you have to write `(-2, signal.shape[axis])` to get the last
        two entries of `axis`, or simply `-1` if you only want to get the last
        entry.

    Shape:
        - Input: :math:`(*, N_\text{in}, *)`
        - Output: :math:`(*, N_\text{out}, *)`, where :math:`N_\text{in}` is
          the input length of the axis to crop and :math:`N_\text{out}` is the
          output length, which is :math:`1` for an integer as `idx` and
          :math:`\text{idx[1]} - \text{idx[0]}` for a tuple with positive
          entries as `idx`.
          :math:`*` can be any additional number of dimensions.

    Example:
        >>> a = np.array([[1, 2], [3, 4]])
        >>> t = Crop(1, axis=1)
        >>> print(t)
        Crop(idx=1, axis=1)
        >>> t(a)
        array([[2],
               [4]])

    """

    def __init__(self, idx, *, axis=-1):
        super().__init__()
        self.idx = idx
        self.axis = axis

    def __call__(self, signal):
        return F.crop(signal, self.idx, axis=self.axis)

    def __repr__(self):
        options = 'idx={0}, axis={1}'.format(self.idx, self.axis)
        return '{0}({1})'.format(self.__class__.__name__, options)


class RandomCrop(object):
    r"""Random crop of specified width along an axis.

    If the signal is too short it is padded by trailing zeros first or
    replicated to fit specified size.

    If the signal is shorter than the desired length, it can be expanded by
    one of these methods:

        * ``'pad'`` expand the signal by adding trailing zeros
        * ``'replicate'`` first replicate the signal so that it matches or
          exceeds the specified size

    * :attr:`size` controls the size of output signal
    * :attr:`method` holds expansion method
    * :attr:`axis` controls axis of cropping
    * :attr:`fix_randomization` controls the randomness

    Args:
        size (int): desired width of spectrogram in samples
        method (str, optional): expansion method. Default: `pad`
        axis (int, optional): axis along to crop. Default: `-1`
        fix_randomization (bool, optional): fix random selection between
            different calls of transform. Default: `False`

    Shape:
        - Input: :math:`(*, N_\text{in}, *)`
        - Output: :math:`(*, N_\text{out}, *)`, where :math:`N_\text{in}` is
          the input length of the axis to crop and :math:`N_\text{out}` is the
          output length as given by `size`.
          :math:`*` can be any additional number of dimensions.

    Example:
        >>> random.seed(0)
        >>> a = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
        >>> t = RandomCrop(2)
        >>> print(t)
        RandomCrop(size=2, method=pad, axis=-1)
        >>> t(a)
        array([[2, 3],
               [6, 7]])

    """

    def __init__(self, size, *, method='pad', axis=-1,
                 fix_randomization=False):
        super().__init__()
        self.size = size
        self.axis = axis
        self.fix_randomization = fix_randomization
        self.expand = Expand(size=0, axis=self.axis, method=method)
        self.idx = [0, 0]
        self.method = method

    def __call__(self, signal):
        input_size = signal.shape[self.axis]
        output_size = self.size

        # Pad or replicate if signal is too short
        if input_size < output_size:
            self.expand.size = output_size
            signal = self.expand(signal)
            input_size = output_size

        # Pad if random crop parameter is fixed and signal is too short
        if self.fix_randomization and input_size < self.idx[1]:
            self.expand.size = self.idx[1]
            signal = self.expand(signal)
            input_size = self.idx[1]

        if not self.fix_randomization:
            self.idx = self.random_index(input_size, output_size)

        return F.crop(signal, self.idx, axis=self.axis)

    @staticmethod
    def random_index(input_size, output_size):
        """Random index for crop.

        Args:
            input_size (int): input signal size
            output_size (int): expected output size

        Returns:
            tuple: random index for cropping

        """
        start = random.randint(0, input_size - output_size)
        return (start, start + output_size)

    def __repr__(self):
        options = 'size={0}, method={1}, axis={2}'.format(
            self.size, self.method, self.axis)
        if self.fix_randomization:
            options += ', fix_randomization=True'
        return '{0}({1})'.format(self.__class__.__name__, options)


class Pad(object):
    r"""Pad along an axis.

    If padding is an integer it pads equally on the left and right of the
    signal. If padding is a tuple with two entries it uses the first for the
    left side and the second for the right side.

    * :attr:`padding` controls the padding to be applied
    * :attr:`value` controls the value used for padding
    * :attr:`axis` controls the axis of padding

    Args:
        padding (int or tuple): padding to apply on the left and right
        value (float, optional): value to pad with. Default: `0`
        axis (int, optional): axis along to pad. Default: `-1`

    Shape:
        - Input: :math:`(*, N_\text{in}, *)`
        - Output: :math:`(*, N_\text{out}, *)`, where :math:`N_\text{in}` is
          the input length of the axis to pad and :math:`N_\text{out} =
          N_\text{in} + \sum \text{padding}` is the output length.
          :math:`*` can be any additional number of dimensions.

    Example:
        >>> a = np.array([[1, 2], [3, 4]])
        >>> t = Pad((0, 1))
        >>> print(t)
        Pad(padding=(0, 1), value=0, axis=-1)
        >>> t(a)
        array([[1, 2, 0],
               [3, 4, 0]])

    """

    def __init__(self, padding, *, value=0, axis=-1):
        super().__init__()
        self.padding = padding
        self.value = value
        self.axis = axis

    def __call__(self, signal):
        return F.pad(signal, self.padding, value=self.value, axis=self.axis)

    def __repr__(self):
        options = ('padding={0}, value={1}, axis={2}'
                   .format(self.padding, self.value, self.axis))
        return '{0}({1})'.format(self.__class__.__name__, options)


class RandomPad(object):
    r"""Random pad along an axis.

    It splits the padding value randomly between the left and right of the
    signal along the specified axis.

    * :attr:`padding` controls the size of padding to be applied
    * :attr:`value` controls the value used for padding
    * :attr:`axis` controls the axis of padding
    * :attr:`fix_randomization` controls the randomness

    Args:
        padding (int): padding to apply randomly split on the left and right
        value (float, optional): value to pad with. Default: `0`
        axis (int, optional): axis along to pad. Default: `-1`
        fix_randomization (bool, optional): fix random selection between
            different calls of transform. Default: `False`

    Shape:
        - Input: :math:`(*, N_\text{in}, *)`
        - Output: :math:`(*, N_\text{out}, *)`, where :math:`N_\text{in}` is
          the input length of the axis to pad and :math:`N_\text{out} =
          N_\text{in} + \sum \text{padding}` is the output length.
          :math:`*` can be any additional number of dimensions.

    Example:
        >>> random.seed(0)
        >>> a = np.array([[1, 2], [3, 4]])
        >>> t = RandomPad(1)
        >>> print(t)
        RandomPad(padding=1, value=0, axis=-1)
        >>> t(a)
        array([[0, 1, 2],
               [0, 3, 4]])

    """

    def __init__(self, padding, *, value=0, axis=-1, fix_randomization=False):
        super().__init__()
        self.padding = padding
        self.value = value
        self.axis = axis
        self.fix_randomization = fix_randomization
        self.pad = None

    def __call__(self, signal):
        if not self.pad or not self.fix_randomization:
            self.pad = self.random_split(self.padding)
        return F.pad(signal, self.pad, value=self.value, axis=self.axis)

    def __repr__(self):
        options = ('padding={0}, value={1}, axis={2}'
                   .format(self.padding, self.value, self.axis))
        if self.fix_randomization:
            options += ', fix_randomization=True'
        return '{0}({1})'.format(self.__class__.__name__, options)

    @staticmethod
    def random_split(number):
        """Split number randomly into two which sum up to number.

        Args:
            number (int): input number to be split

        Returns:
            tuple: randomly splitted number

        """
        left = random.randint(0, number)
        return (left, number - left)


class Replicate(object):
    r"""Replicate along an axis.

    * :attr:`repetitions` controls number of signal replications
    * :attr:`axis` controls the axis of replication

    Args:
        repetitions (int or tuple): number of times to replicate signal
        axis (int, optional): axis along which to replicate. Default: `-1`

    Shape:
        - Input: :math:`(*, N_\text{in}, *)`
        - Output: :math:`(*, N_\text{out}, *)`, where :math:`N_\text{in}` is
          the input length of the axis to replicate and :math:`N_\text{out} =
          N_\text{in} \cdot \text{repetitions}` is the output length.
          :math:`*` can be any additional number of dimensions.

    Example:
        >>> a = np.array([[1, 2, 3]])
        >>> t = Replicate(3)
        >>> print(t)
        Replicate(repetitions=3, axis=-1)
        >>> t(a)
        array([[1, 2, 3, 1, 2, 3, 1, 2, 3]])

    """

    def __init__(self, repetitions, *, axis=-1):
        super().__init__()
        self.repetitions = repetitions
        self.axis = axis

    def __call__(self, signal):
        return F.replicate(signal, self.repetitions, axis=self.axis)

    def __repr__(self):
        options = ('repetitions={0}, axis={1}'
                   .format(self.repetitions, self.axis))
        return '{0}({1})'.format(self.__class__.__name__, options)


class RandomReplicate(object):
    r"""Replicate by a random number of times along an axis.

    * :attr:`repetitions` holds number of times to replicate signal
    * :attr:`axis` controls the axis of replication
    * :attr:`fix_randomization` controls the randomness

    Args:
        max_repetitions (int, optional): controls the maximum number of times
            a signal is allowed to be replicated. Default: `100`
        axis (int, optional): axis along which to pad. Default: `-1`
        fix_randomization (bool, optional): fix random selection between
            different calls of transform. Default: `False`

    Shape:
        - Input: :math:`(*, N_\text{in}, *)`
        - Output: :math:`(*, N_\text{out}, *)`, where :math:`N_\text{in}` is
          the input length of the axis to pad and :math:`N_\text{out} =
          N_\text{in} \cdot \text{repetitions}` is the output length.
          :math:`*` can be any additional number of dimensions.

    Example:
        >>> random.seed(0)
        >>> a = np.array([1, 2, 3])
        >>> t = RandomReplicate(max_repetitions=3)
        >>> print(t)
        RandomReplicate(max_repetitions=3, repetitions=None, axis=-1)
        >>> t(a)
        array([1, 2, 3, 1, 2, 3, 1, 2, 3])

    """

    def __init__(self, *, max_repetitions=100, axis=-1,
                 fix_randomization=False):
        super().__init__()
        self.repetitions = None
        self.max_repetitions = max_repetitions
        self.axis = axis
        self.fix_randomization = fix_randomization

    def __call__(self, signal):
        if self.repetitions is None or not self.fix_randomization:
            self.repetitions = random.randint(0, self.max_repetitions)
        return F.replicate(signal, self.repetitions, axis=self.axis)

    def __repr__(self):
        options = ('max_repetitions={0}, repetitions={1}, axis={2}'
                   .format(self.max_repetitions, self.repetitions, self.axis))
        if self.fix_randomization:
            options += ', fix_randomization=True'
        return '{0}({1})'.format(self.__class__.__name__, options)


class Expand(object):
    r"""Expand signal.

    Ensures that the signal matches the desired output size by padding or
    replicating it.

    * :attr:`size` controls the size of output signal
    * :attr:`method` controls whether to replicate signal or pad it
    * :attr:`axis` controls axis of expansion

    The expansion is done by one of these methods:

        * ``'pad'`` expand the signal by adding trailing zeros
        * ``'replicate'`` replicate the signal to match the specified size.
          If result exceeds specified size after replication, the signal will
          then be cropped

    Args:
        size (int): desired length of output signal in samples
        method (str, optional): expansion method. Default: `pad`
        axis (int, optional): axis along to crop. Default: `-1`

    Shape:
        - Input: :math:`(*, N_\text{in}, *)`
        - Output: :math:`(*, N_\text{out}, *)`, where :math:`N_\text{in}` is
          the input length of the axis to expand and :math:`N_\text{out}` is
          the output length as given by `size`.
          :math:`*` can be any additional number of dimensions.

    Example:
        >>> a = np.array([[1, 2, 3]])
        >>> t = Expand(6)
        >>> print(t)
        Expand(size=6, method=pad, axis=-1)
        >>> t(a)
        array([[1, 2, 3, 0, 0, 0]])

    """

    def __init__(self, size, *, method='pad', axis=-1):
        super().__init__()
        self.size = size
        self.method = method
        self.axis = axis

    def __call__(self, signal):
        input_size = signal.shape[self.axis]
        output_size = self.size

        if input_size < output_size:
            if self.method == 'replicate':
                signal = F.replicate(signal, output_size // input_size + 1,
                                     axis=self.axis)
                signal = F.crop(signal, (0, output_size), axis=self.axis)
            elif self.method == 'pad':
                signal = F.pad(signal, (0, output_size - input_size),
                               axis=self.axis)
        return signal

    def __repr__(self):
        options = 'size={0}, method={1}, axis={2}'.format(
            self.size, self.method, self.axis)
        return '{0}({1})'.format(self.__class__.__name__, options)


class Downmix(object):
    r"""Downmix to the provided number of channels.

    The downmix is done by one of these methods:

        * ``'mean'`` replace last desired channel by mean across itself and
          all remaining channels
        * ``'crop'`` drop all remaining channels

    * :attr:`channels` controls the number of desired channels
    * :attr:`method` controls downmixing method
    * :attr:`axis` controls axis of downmix

    Args:
        channels (int): number of desired channels
        method (str, optional): downmix method. Default: `'mean'`
        axis (int, optional): axis to downmix. Default: `-2`

    Shape:
        - Input: :math:`(*, C_\text{in}, *)`
        - Output: :math:`(*, C_\text{out}, *)`, where :math:`C_\text{in}` is
          the number of input channels and :math:`C_\text{out}` is the number
          of output channels as given by `channels`. :math:`*` can be any
          additional number of dimensions.

    Example:
        >>> a = np.array([[1, 2], [3, 4]])
        >>> t = Downmix(1, axis=0)
        >>> print(t)
        Downmix(channels=1, method=mean, axis=0)
        >>> t(a)
        array([[2, 3]])

    """

    def __init__(self, channels, *, method='mean', axis=-2):
        super().__init__()
        self.channels = channels
        self.method = method
        self.axis = axis

    def __call__(self, signal):
        return F.downmix(signal, self.channels, method=self.method,
                         axis=self.axis)

    def __repr__(self):
        options = ('channels={0}, method={1}, axis={2}'
                   .format(self.channels, self.method, self.axis))
        return '{0}({1})'.format(self.__class__.__name__, options)


class Upmix(object):
    r"""Upmix to the provided number of channels.

    The upmix is achieved by adding the same signal in the additional channels.
    This signal is calculated by one of the following methods:

        * ``'mean'`` mean across all input channels
        * ``'zero'`` zeros
        * ``'repeat'`` last input channel

    * :attr:`channels` controls the number of desired channels
    * :attr:`method` controls downmixing method
    * :attr:`axis` controls axis of upmix

    Args:
        channels (int): number of desired channels
        method (str, optional): upmix method. Default: `'mean'`
        axis (int, optional): axis to upmix. Default: `-2`

    Shape:
        - Input: :math:`(*, C_\text{in}, *)`
        - Output: :math:`(*, C_\text{out}, *)`, where :math:`C_\text{in}` is
          the number of input channels and :math:`C_\text{out}` is the number
          of output channels as given by `channels`.
          :math:`*` can be any additional number of dimensions.

    Example:
        >>> a = np.array([[1, 2], [3, 4]])
        >>> t = Upmix(3, axis=0)
        >>> print(t)
        Upmix(channels=3, method=mean, axis=0)
        >>> t(a)
        array([[1., 2.],
               [3., 4.],
               [2., 3.]])

    """

    def __init__(self, channels, *, method='mean', axis=-2):
        super().__init__()
        self.channels = channels
        self.method = method
        self.axis = axis

    def __call__(self, signal):
        return F.upmix(signal, self.channels, method=self.method,
                       axis=self.axis)

    def __repr__(self):
        options = ('channels={0}, method={1}, axis={2}'
                   .format(self.channels, self.method, self.axis))
        return '{0}({1})'.format(self.__class__.__name__, options)


class Remix(object):
    r"""Remix to the provided number of channels.

    The remix is achieved by repeating the mean of all other channels or by
    replacing the last desired channel by the mean across all channels.

    It is internally achieved by running :class:`Upmix` or :class:`Downmix`
    with method `mean`.

    * :attr:`channels` controls the number of desired channels
    * :attr:`axis` controls axis of upmix

    Args:
        channels (int): number of desired channels
        axis (int, optional): axis to upmix. Default: `-2`

    Shape:
        - Input: :math:`(*, C_\text{in}, *)`
        - Output: :math:`(*, C_\text{out}, *)`, where :math:`C_\text{in}` is
          the number of input channels and :math:`C_\text{out}` is the number
          of output channels as given by `channels`.
          :math:`*` can be any additional number of dimensions.

    Example:
        >>> a = np.array([[1, 2], [3, 4]])
        >>> t = Remix(3, axis=0)
        >>> print(t)
        Remix(channels=3, axis=0)
        >>> t(a)
        array([[1., 2.],
               [3., 4.],
               [2., 3.]])

    """

    def __init__(self, channels, *, method='mean', axis=-2):
        super().__init__()
        self.channels = channels
        self.axis = axis

    def __call__(self, signal):
        if self.channels > np.atleast_2d(signal).shape[self.axis]:
            signal = F.upmix(signal, self.channels, method='mean',
                             axis=self.axis)
        elif self.channels < np.atleast_2d(signal).shape[self.axis]:
            signal = F.downmix(signal, self.channels, method='mean',
                               axis=self.axis)
        return signal

    def __repr__(self):
        options = 'channels={}, axis={}'.format(self.channels, self.axis)
        return '{}({})'.format(self.__class__.__name__, options)


class Normalize(object):
    r"""Normalize signal.

    Ensure the maximum of the absolute value of the signal is 1.

    * :attr:`axis` controls axis for normalization

    Args:
        axis (int, optional): axis for normalization. Default: `-1`

    Shape:
        - Input: :math:`(*)`
        - Output: :math:`(*)`, where :math:`*` can be any number of dimensions.

    Example:
        >>> a = np.array([1, 2, 3, 4])
        >>> t = Normalize()
        >>> print(t)
        Normalize(axis=-1)
        >>> t(a)
        array([0.25, 0.5 , 0.75, 1.  ])

    """

    def __init__(self, *, axis=-1):
        super().__init__()
        self.axis = axis

    def __call__(self, signal):
        return F.normalize(signal, axis=self.axis)

    def __repr__(self):
        options = 'axis={0}'.format(self.axis)
        return '{0}({1})'.format(self.__class__.__name__, options)


class Standardize(object):
    r"""Standardize signal.

    Ensure the signal has a mean value of 0 and a variance of 1.

    * :attr:`mean` controls whether mean centering will be applied
    * :attr:`std` controls whether standard deviation normalization will be
      applied
    * :attr:`axis` controls axis for standardization

    Args:
        mean (bool, optional): apply mean centering. Default: `True`
        std (bool, optional): normalize by standard deviation. Default: `True`
        axis (int, optional): standardize only along the given axis.
            Default: `-1`

    Shape:
        - Input: :math:`(*)`
        - Output: :math:`(*)`, where :math:`*` can be any number of dimensions.

    Example:
        >>> a = np.array([1, 2, 3, 4])
        >>> t = Standardize()
        >>> print(t)
        Standardize(axis=-1, mean=True, std=True)
        >>> t(a)
        array([-1.34164079, -0.4472136 ,  0.4472136 ,  1.34164079])

    """

    def __init__(self, *, mean=True, std=True, axis=-1):
        super().__init__()
        self.axis = axis
        self.mean = mean
        self.std = std

    def __call__(self, signal):
        return F.standardize(signal, axis=self.axis,
                             mean=self.mean, std=self.std)

    def __repr__(self):
        options = 'axis={0}'.format(self.axis)
        if self.mean:
            options += ', mean=True'
        if self.std:
            options += ', std=True'
        return '{0}({1})'.format(self.__class__.__name__, options)


class Resample(object):
    r"""Resample to new sampling rate.

    The signal is resampled by one of the following methods.

        * ``'kaiser_best'`` as implemented by resampy
        * ``'kaiser_fast'`` as implemented by resampy
        * ``'scipy'`` uses scipy for resampling

    * :attr:`input_sampling_rate` controls input sample rate in Hz
    * :attr:`output_sampling_rate` controls output sample rate in Hz
    * :attr:`method` controls the resample method
    * :attr:`axis` controls axis for resampling

    Args:
        input_sampling_rate (int): input sample rate in Hz
        output_sampling_rate (int): output sample rate in Hz
        method (str, optional): resample method. Default: `kaiser_best`
        axis (int, optional): axis for resampling. Default: `-1`

    Note:
        If the default method `kaiser_best` is too slow for your purposes,
        you should try `scipy` instead. `scipy` is the fastest method, but
        might crash for very long signals.

    Shape:
        - Input: :math:`(*)`
        - Output: :math:`(*)`, where :math:`*` can be any number of dimensions.

    Example:
        >>> a = np.array([1, 2, 3, 4])
        >>> t = Resample(4, 2)
        >>> print(t)
        Resample(input_sampling_rate=4, output_sampling_rate=2, method=kaiser_best, axis=-1)
        >>> t(a)
        array([0, 2])

    """  # noqa: E501

    def __init__(self, input_sampling_rate, output_sampling_rate, *,
                 method='kaiser_best', axis=-1):
        super().__init__()
        self.input_sampling_rate = input_sampling_rate
        self.output_sampling_rate = output_sampling_rate
        self.method = method
        self.axis = axis

    def __call__(self, signal):
        if self.method == 'scipy':
            out_samples = int(signal.shape[self.axis]
                              * self.output_sampling_rate
                              / float(self.input_sampling_rate))
            signal = scipy.signal.resample(signal, out_samples, axis=self.axis)
        else:
            signal = resampy.resample(signal, self.input_sampling_rate,
                                      self.output_sampling_rate,
                                      method=self.method, axis=self.axis)
        return signal

    def __repr__(self):
        options = ('input_sampling_rate={0}, output_sampling_rate={1}, '
                   'method={2}, axis={3}'
                   .format(self.input_sampling_rate, self.output_sampling_rate,
                           self.method, self.axis))
        return '{0}({1})'.format(self.__class__.__name__, options)


class Spectrogram(object):
    r"""Spectrogram of an audio signal.

    The spectrogram is calculated by librosa and its magnitude is returned as
    real valued matrix.

    * :attr:`window_size` controls FFT window size in samples
    * :attr:`hop_size` controls STFT window hop size in samples
    * :attr:`fft_size` controls number of frequency bins in STFT
    * :attr:`window` controls window function of spectrogram computation
    * :attr:`axis` controls axis of spectrogram computation
    * :attr:`phase` holds the phase of the spectrogram

    Args:
        window_size (int): size of STFT window in samples
        hop_size (int): size of STFT window hop in samples
        fft_size(int, optional): number of frequency bins in STFT. If `None`,
            then it defaults to `window_size`. Default: `None`
        window (str, tuple, number, function, or numpy.ndarray, optional): type
            of STFT window. Default: `hann`
        axis (int, optional): axis of STFT calculation. Default: `-1`

    Shape:
        - Input: :math:`(*, N_\text{in}, *)`
        - Output: :math:`(*, N_f, N_t, *)`, where :math:`N_\text{in}` is
          the number of input samples and
          :math:`N_f = {\text{window\_size} \over 2} + 1` is the number of
          output samples along the frequency axis of the spectrogram, and
          :math:`N_t = \lceil {1 \over \text{hop\_size}} (N_\text{in} +
          {\text{window\_size} \over 2}) \rceil` is the number of output
          samples along the time axis of the spectrogram.
          :math:`*` can be any additional number of dimensions.

    Example:
        >>> a = np.array([1., 2., 3., 4.])
        >>> t = Spectrogram(2, 2)
        >>> print(t)
        Spectrogram(window_size=2, hop_size=2, axis=-1)
        >>> t(a)
        array([[1., 3., 3.],
               [1., 3., 3.]], dtype=float32)

    """

    def __init__(self, window_size, hop_size, *, fft_size=None,
                 window='hann', axis=-1):
        super().__init__()
        self.window_size = window_size
        self.hop_size = hop_size
        self.fft_size = fft_size
        self.window = window
        self.axis = axis
        self.phase = []

    def __call__(self, signal):
        spectrogram = F.stft(signal, self.window_size, self.hop_size,
                             fft_size=self.fft_size, window=self.window,
                             axis=self.axis)
        magnitude, self.phase = librosa.magphase(spectrogram)
        return magnitude

    def __repr__(self):
        options = ('window_size={0}, hop_size={1}, axis={2}'
                   .format(self.window_size, self.hop_size, self.axis))
        if self.window != 'hann':
            options += ', window={0}'.format(self.window)
        return '{0}({1})'.format(self.__class__.__name__, options)


class Log(object):
    r"""Logarithmic transform of an input signal.

    * :attr:`magnitude_boost` controls the non-negative value added to the
      magnitude of the signal before applying the logarithmus

    Args:
        magnitude_boost (float, optional): positive value added to the
            magnitude of the signal before applying the logarithmus. Default:
            `1e-7`

    Shape:
        - Input: :math:`(*)`
        - Output: :math:`(*)`, where :math:`*` can be any additional number of
          dimensions.

    Example:
        >>> a = np.array([1., 2., 3., 4.])
        >>> spect = Spectrogram(window_size=2, hop_size=2)
        >>> t = Log()
        >>> print(t)
        Log(magnitude_boost=1e-07)
        >>> t(spect(a))
        array([[1.1920928e-07, 1.0986123e+00, 1.0986123e+00],
               [1.1920928e-07, 1.0986123e+00, 1.0986123e+00]], dtype=float32)

    """

    def __init__(self, magnitude_boost=1e-7):
        self.magnitude_boost = magnitude_boost

        if self.magnitude_boost < 0:
            raise ValueError('`magnitude_boost` has to be >=0, but is {}'
                             .format(self.magnitude_boost))

    def __call__(self, signal):
        signal = np.log(signal + self.magnitude_boost)
        return signal

    def __repr__(self):
        options = ('magnitude_boost={}'.format(self.magnitude_boost))
        return '{0}({1})'.format(self.__class__.__name__, options)


class LogSpectrogram(object):
    r"""Logarithmic spectrogram of an audio signal.

    The spectrogram is calculated by librosa and its magnitude is returned as
    real valued matrix. If `normalize` is set to `True` the magnitude is
    z-normalized by subtracting its mean and dividing by its standard
    deviation.

    * :attr:`window_size` controls FFT window size in samples
    * :attr:`hop_size` controls STFT window hop size in samples
    * :attr:`window` controls window function of spectrogram computation
    * :attr:`normalize` controls normalization of spectrogram
    * :attr:`magnitude_boost` controls the positive value added to the
      magnitude of the spectrogram before applying the logarithmus
    * :attr:`axis` controls axis of spectrogram computation
    * :attr:`phase` holds the phase of the spectrogram

    Args:
        window_size (int): size of STFT window in samples
        hop_size (int): size of STFT window hop in samples
        window (str, tuple, number, function, or numpy.ndarray, optional): type
            of STFT window. Default: `hann`
        normalize (bool, optional): normalize spectrogram. Default: `False`
        magnitude_boost (float, optional): positive value added to the
            magnitude of the spectrogram before applying the logarithmus.
            Default: `1e-7`
        axis (int, optional): axis of STFT calculation. Default: `-1`

    Shape:
        - Input: :math:`(*, N_\text{in}, *)`
        - Output: :math:`(*, N_f, N_t, *)`, where :math:`N_\text{in}` is
          the number of input samples and
          :math:`N_f = 1 + {\text{window\_size} \over 2}` is the number of
          output samples along the frequency axis of the spectrogram, and
          :math:`N_t = \lceil {1 \over \text{hop\_size}} (N_\text{in} +
          {\text{window\_size} \over 2}) \rceil` is the number of output
          samples along the time axis of the spectrogram.
          :math:`*` can be any additional number of dimensions.

    Example:
        >>> a = np.array([1., 2., 3., 4.])
        >>> t = LogSpectrogram(2, 2)
        >>> print(t)
        LogSpectrogram(window_size=2, hop_size=2, magnitude_boost=1e-07, axis=-1, normalize=False)
        >>> t(a)
        array([[1.1920928e-07, 1.0986123e+00, 1.0986123e+00],
               [1.1920928e-07, 1.0986123e+00, 1.0986123e+00]], dtype=float32)

    """  # noqa: E501

    def __init__(self, window_size, hop_size, *, window='hann',
                 normalize=False, magnitude_boost=1e-7, axis=-1):
        super().__init__()
        warn('LogSpectrogram will be removed in a future version. '
             'Please use a Compose transform consisting of Spectrogram and '
             'Log instead.', DeprecationWarning)
        self.window_size = window_size
        self.hop_size = hop_size
        self.window = window
        self.normalize = normalize
        self.magnitude_boost = magnitude_boost
        self.axis = axis
        self.phase = []
        if self.normalize:
            warn('Spectrogram normalization is enabled. Do not combine it '
                 'with signal reconstruction as it will degrade the result.',
                 UserWarning)
        if self.magnitude_boost <= 0:
            raise ValueError('`magnitude_boost` has to be >0, but is {}'
                             .format(self.magnitude_boost))

    def __call__(self, signal):
        spectrogram = F.stft(signal, self.window_size, self.hop_size,
                             window=self.window, axis=self.axis)
        magnitude, self.phase = librosa.magphase(spectrogram)
        magnitude = np.log(magnitude + self.magnitude_boost)

        if self.normalize:
            magnitude -= magnitude.mean()
            magnitude /= magnitude.std()

        return magnitude

    def __repr__(self):
        options = ('window_size={}, hop_size={}, magnitude_boost={}, axis={}'
                   .format(self.window_size, self.hop_size,
                           self.magnitude_boost, self.axis))
        if not self.normalize:
            options += ', normalize=False'
        if self.window != 'hann':
            options += ', window={0}'.format(self.window)
        return '{0}({1})'.format(self.__class__.__name__, options)


class RandomAdditiveMix(object):
    r"""Mix two signals additively by a randomly picked ratio.

    Randomly pick a signal from an augmentation data set and mix it with the
    actual signal by a signal-to-noise ratio in dB randomly selected from a
    list of possible ratios.

    The signal from the augmentation data set is expanded, cropped,
    or has its number of channels adjusted by a downmix or upmix using
    :class:`Remix` if necessary.

    The signal can be expanded by:

        * ``'multiple'`` loading multiple files from the
          augmentation data set and concatenating them along the time axis
        * ``'pad'`` expand the signal by adding trailing zeros
        * ``'replicate'`` replicate the signal to match the specified size.
          If result exceeds specified size after replication, the signal will
          then be cropped

    The signal can be cropped by:

        * ``'start'`` crop signal from the beginning of the file all
          the way to the necessary length
        * ``'random'`` starts at a random offset from the beginning of the file

    * :attr:`dataset` controls the data set used for augmentation
    * :attr:`ratio` controls the ratio in dB between mixed signals
    * :attr:`ratios` controls the ratios to be randomly picked from
    * :attr:`normalize` controls if the mixed signal is normalized
    * :attr:`expand_method` controls if the signal from the augmented data
      set is automatically expanded according to an expansion rule.
      Default: `pad`
    * :attr:`crop_method` controls how the signal is cropped. Is only
      relevant if the augmentation signal is longer than the input one,
      or if `expand_method` is set to `multiple`. Default: `random`
    * :attr:`percentage_silence` controls the percentage of the input data
      that will be mixed with silence. Should be between `0` and `1`.
      Default: `1`
    * :attr:`time_axis` controls time axis for automatic signal adjustment
    * :attr:`channel_axis` controls channel axis for automatic signal
      adjustment
    * :attr:`fix_randomization` controls the randomness of the ratio selection

    Note:
        :attr:`fix_randomization` covers only the selection of the ratio. The
        selection of a signal from the augmentation data set and its signal
        length adjustment will always be random.

    Args:
        dataset (torch.utils.data.Dataset): data set for augmentation
        ratios (list of int, optional): mix ratios in dB to randomly pick from
            (e.g. SNRs). Default: `[0, 15, 30]`
        normalize (bool, optional): normalize mixture. Default: `False`
        expand_method (str, optional): controls the adjustment of
            the length data set that is added to the original data set.
            Default: `pad`
        crop_method (str, optional): controls the crop transform that will be
            called on the mix signal if it is longer than the input signal.
            Default: `random`
        percentage_silence (float, optional): controls the percentage of
            input data that should be augmented with silence. Default: `0`
        time_axis (int, optional): length axis of both data sets. Default: `-1`
        channel_axis (int, optional): channels axis of both data sets.
            Default: `-2`
        fix_randomization (bool, optional): freeze random selection between
            different calls of transform. Default: `False`

    Shape:
        - Input: :math:`(*, C, N, *)`
        - Output: :math:`(*, C, N, *)`, where :math:`C` is the number of
          channels and :math:`N` is the number of samples. They don't have to
          be placed in the order shown here, but the order is preserved during
          transformation.
          :math:`*` can be any additional number of dimensions.

    Example:
        >>> from audtorch import datasets
        >>> np.random.seed(0)
        >>> a = np.array([[1, 2], [3, 4]])
        >>> noise = datasets.WhiteNoise(duration=1, sampling_rate=2)
        >>> t = RandomAdditiveMix(noise, ratios=[3], expand_method='pad')
        >>> print(t)
        RandomAdditiveMix(dataset=WhiteNoise, ratios=[3], ratio=None, percentage_silence=0, expand_method=pad, crop_method=random, time_axis=-1, channel_axis=-2)
        >>> t(a)
        array([[3.67392992, 2.60655362],
               [5.67392992, 4.60655362]])

    """  # noqa: E501

    def __init__(self, dataset, *, ratios=[0, 15, 30], normalize=False,
                 expand_method='pad', crop_method='random',
                 percentage_silence=0, time_axis=-1, channel_axis=-2,
                 fix_randomization=False):
        super().__init__()
        self.dataset = dataset
        self.ratios = ratios
        self.ratio = None
        self.percentage_silence = percentage_silence
        self.normalize = normalize
        self.expand_method = expand_method
        self.crop_method = crop_method
        self.time_axis = time_axis
        self.channel_axis = channel_axis
        self.fix_randomization = fix_randomization

        self._remix = Remix(0, axis=channel_axis)
        if self.expand_method != 'multiple':
            self._expand = Expand(0, axis=self.time_axis,
                                  method=self.expand_method)

    def __call__(self, signal1):
        if not self.fix_randomization or not self.ratio:
            self.ratio = random.choice(self.ratios)

        if random.random() < self.percentage_silence:
            signal2 = np.zeros(signal1.shape)
        else:
            signal2 = random.choice(self.dataset)[0]

            samples_signal1 = signal1.shape[self.time_axis]
            samples_signal2 = signal2.shape[self.time_axis]
            channels_signal1 = np.atleast_2d(signal1).shape[self.channel_axis]
            channels_signal2 = np.atleast_2d(signal2).shape[self.channel_axis]

            # Extend too short signal2
            if samples_signal1 > samples_signal2:
                if self.expand_method == 'multiple':
                    self._remix.channels = channels_signal2
                    while samples_signal2 < samples_signal1:
                        new_signal = random.choice(self.dataset)[0]
                        new_signal = self._remix(new_signal)
                        signal2 = np.concatenate((signal2, new_signal),
                                                 axis=self.time_axis)
                        samples_signal2 = signal2.shape[self.time_axis]
                else:
                    self._expand.size = samples_signal1
                    signal2 = self._expand(signal2)
                    samples_signal2 = signal2.shape[self.time_axis]

            # Crop too long signal2 to match signal1
            if samples_signal1 < samples_signal2:
                if self.crop_method == 'random':
                    idx = RandomCrop.random_index(
                        samples_signal2, samples_signal1
                    )
                elif self.crop_method == 'start':
                    idx = (0, samples_signal1)
                signal2 = F.crop(signal2, idx, axis=self.time_axis)

            # Adjust number of channels of signal2 to signal1
            if channels_signal1 != channels_signal2:
                self._remix.channels = channels_signal1
                signal2 = self._remix(signal2)

        mixture = F.additive_mix(signal1, signal2, self.ratio)

        if self.normalize:
            mixture = F.normalize(mixture)

        return mixture

    def __repr__(self):
        options = ('dataset={0}, ratios={1}, ratio={2}, '
                   'percentage_silence={3}, '
                   'expand_method={4}, crop_method={5}, '
                   'time_axis={6}, channel_axis={7}'
                   .format(self.dataset.__class__.__name__, self.ratios,
                           self.ratio, self.percentage_silence,
                           self.expand_method, self.crop_method,
                           self.time_axis, self.channel_axis))
        if self.normalize:
            options += ', normalize=True'
        if self.fix_randomization:
            options += ', fix_randomization=True'
        return '{0}({1})'.format(self.__class__.__name__, options)
